import tkinter as tk

from octopus_python_client.common import Common
from octopus_python_client.config import Config, SystemConfig
from octopus_python_client.gui.wizard import Wizard


class MainGUI:
    def __init__(self, width: int = 800, height: int = 600):
        self.width = width
        self.height = height
        self.config = Config()
        self.server = Common(config=self.config)
        self.source_config = Config(is_source_server=True)
        self.source = Common(config=self.source_config)

    def set_gui(self):
        window = tk.Tk()
        window.title(SystemConfig.TITLE)
        window.option_add("*font", "calibri 14")
        window["bg"] = "black"

        # window.geometry('800x600')

        wiz = Wizard(window, server=self.server, source=self.source)
        wiz.pack()

        window.mainloop()


if __name__ == "__main__":
    MainGUI().set_gui()



from octopus_python_client.constants import Constants


class Actions:
    ACTION_CLONE = "clone"  # clone a new item
    ACTION_CLONE_CHILD = "clone_child"  # clone a child item
    ACTION_CLONE_PROCESS_STEP = "clone_process_step"  # clone a project process step
    ACTION_CLONE_PROJECT = "clone_project"  # clone a project including process
    ACTION_CLONE_PROJECT_RELATED = "clone_project_related"  # clone a project related items from a project to projects
    ACTION_CLONE_SERVER = "clone_server"  # clone a few types from one server to another server
    ACTION_CLONE_SPACE = "clone_space"  # clone a few types from one space to another space
    ACTION_CLONE_SPACE_ITEM = "clone_space_item"  # clone Octopus single item from one space to another space
    ACTION_CREATE = "create"  # create a new item
    ACTION_CREATE_DEPLOYMENT = "create_deployment"
    ACTION_CREATE_RELEASE = "create_release"
    ACTION_CREATE_RELEASE_DEPLOYMENT = "create_release_deployment"
    ACTION_DELETE = "delete"  # delete an item
    ACTION_DELETE_PROCESS_STEP = "delete_process_step"  # delete a project process step
    ACTION_DELETE_PROJECT = "delete_project"  # delete a project including process
    ACTION_DELETE_PROJECTS = "delete_projects"  # delete projects inside project groups and excluding projects
    ACTION_DELETE_TYPE = "delete_type"  # delete all items_dict under one type
    ACTION_DELETE_TYPES = "delete_types"  # delete all items_dict under specified types (or delete all types)
    ACTION_GET = "get"  # get one item
    ACTION_GET_CHILD = "get_child"  # get a child item
    ACTION_GET_PROJECT = "get_project"  # get a project including process
    ACTION_GET_SPACES = "get_spaces"  # get multiple spaces
    ACTION_GET_TYPE = "get_type"  # get all items_dict under one type
    ACTION_GET_TYPES = "get_types"  # get all types
    ACTION_GUI = "gui"  # start Octopus python client GUI
    ACTION_PROJECT_UPDATE_VARIABLE_SETS = "update_variable_sets"  # update the variable sets for a project
    ACTION_TASK_STATUS = "task_status"
    ACTION_UPDATE = "update"  # update one item
    ACTION_UPDATE_CHILD = "update_child"  # update a child item
    ACTION_UPDATE_MERGE = "update_merge"  # update one item and merge the existing sub-items_dict
    ACTION_WAIT_TASK = "wait_task"


ACTIONS_DICT = {
    # Actions.ACTION_GET_SPACES: "download the data of multiple types in a number of spaces or all spaces",
    # Actions.ACTION_GET_TYPES: "download the data of multiple types in one space",
    Actions.ACTION_CLONE_SPACE: "clone multiple types from one space to another space on the same server or the "
                                "different server",
    Actions.ACTION_CLONE_SPACE_ITEM: "clone one item of a specific type within a space, to another space, or to another"
                                     " server",
    Actions.ACTION_CLONE_PROJECT_RELATED: f"clone project related items {Constants.PROJECT_RELATED_TYPES} from one "
                                          f"project to other projects",
    Actions.ACTION_CREATE_RELEASE: "create a new release for a project",
    Actions.ACTION_CREATE_DEPLOYMENT: "create a new deployment for a release of a project",
    Actions.ACTION_GET_SPACES: "download data and configurations from one or more Octopus spaces to local files",
    Actions.ACTION_GET: "download single item data/configuration from Octopus server to a local file",
    Actions.ACTION_UPDATE: "update single item data/configuration on Octopus server from a local file"
}

MIGRATION_LIST = [Actions.ACTION_CLONE_PROJECT_RELATED, Actions.ACTION_CLONE_SPACE_ITEM, Actions.ACTION_CLONE_SPACE,
                  Actions.ACTION_CLONE_SERVER]



import tkinter as tk

from octopus_python_client.common import Common
from octopus_python_client.gui.actions_widgets import ActionsWidgets
from octopus_python_client.gui.options_widgets import OptionsWidgets
from octopus_python_client.gui.servers_widgets import ServersWidgets
from octopus_python_client.gui.spaces_widgets import SpacesWidgets
from octopus_python_client.gui.submit_widgets import SubmitWidgets


class Wizard(tk.Frame):
    def __init__(self, parent, server: Common, source: Common):
        super().__init__(parent)

        self.server = server
        self.source = source

        self.current_step_index = None
        self.current_step_widgets = None
        self.steps = [ActionsWidgets, ServersWidgets, SpacesWidgets, OptionsWidgets, SubmitWidgets]

        self.button_frame = tk.Frame(self, bd=1, relief="raised")
        self.back_button = tk.Button(self.button_frame, text="<< Back", command=self.back)
        self.next_button = tk.Button(self.button_frame, text="Next >>", command=self.next)
        self.submit_button = tk.Button(self.button_frame, text="Submit", command=self.submit)
        self.restart_button = tk.Button(self.button_frame, text="<<<< Restart", command=self.restart)
        self.button_frame.pack(side="bottom", fill="x")

        self.content_frame = tk.Frame(self)
        self.show_step(0)
        self.content_frame.pack(side="top", fill="both", expand=True)

    def show_step(self, step):

        if self.current_step_widgets is not None:
            # remove current step
            self.current_step_widgets.pack_forget()

        self.current_step_index = step
        self.next_button.config(state=tk.NORMAL)
        self.submit_button.config(state=tk.NORMAL)
        self.current_step_widgets = self.steps[step](parent=self.content_frame, server=self.server, source=self.source,
                                                     next_button=self.next_button, submit_button=self.submit_button)
        self.current_step_widgets.pack(fill="both", expand=True)

        if step == 0:
            # first step
            self.back_button.pack_forget()
            self.restart_button.pack_forget()
            self.next_button.pack(side="right")
            self.submit_button.pack_forget()

        elif step == len(self.steps) - 1:
            # last step
            self.restart_button.pack(side="left")
            self.back_button.pack(side="left")
            self.next_button.pack_forget()
            self.submit_button.pack(side="right")

        else:
            # all other steps
            self.restart_button.pack(side="left")
            self.back_button.pack(side="left")
            self.next_button.pack(side="right")
            self.submit_button.pack_forget()

    def next(self):
        if self.current_step_widgets.process_config():
            self.show_step(self.current_step_index + 1)

    def back(self):
        if self.current_step_widgets.process_config():
            self.show_step(self.current_step_index - 1)

    def submit(self):
        self.current_step_widgets.process_config()
        self.current_step_widgets.start_run()

    def restart(self):
        self.current_step_widgets.process_config()
        self.show_step(0)



import copy
import logging
import tkinter as tk
from tkinter import ttk

from octopus_python_client.common import id_key, name_key
from octopus_python_client.utilities.helper import find_item

logger = logging.getLogger(__name__)


class CommonWidgets:
    COMBOBOX_VALUES_KEY = "values"
    HEIGHT_7 = 7
    READ_ONLY = "readonly"
    ROW_0 = 0
    ROW_1 = 1
    ROW_SIZE = 5
    SELECTED = "1"
    SPAN_1 = 1
    SPAN_2 = 2
    SPAN_3 = 3
    UNSELECTED = "0"
    WIDTH_10 = 10
    WIDTH_20 = 20
    WIDTH_40 = 40
    WIDTH_80 = 80
    WIDTH_120 = 120

    # this is for list of strings
    @staticmethod
    def set_radio_names_frame(parent, list_names: list, default_name: str = None, title="Select one: "):
        names_frame = tk.Frame(parent)
        tk.Label(names_frame, text=title, bd=2) \
            .grid(row=CommonWidgets.ROW_0, columnspan=CommonWidgets.ROW_SIZE, sticky=tk.W)
        name_var = tk.StringVar()
        list_names_copy = copy.deepcopy(list_names)
        list_names_copy.sort()
        if default_name and default_name in list_names_copy:
            name_var.set(default_name)
        else:
            name_var.set(list_names_copy[0])
        for index, name in enumerate(list_names_copy):
            tk.Radiobutton(names_frame, text=f"{name}", variable=name_var, value=name, justify=tk.LEFT,
                           command=lambda *args: None) \
                .grid(row=CommonWidgets.ROW_1 + int(index / CommonWidgets.ROW_SIZE),
                      column=index % CommonWidgets.ROW_SIZE, sticky=tk.W)  # , columnspan=1
        names_frame.grid(sticky=tk.W)
        return name_var

    # this is for items like Id: Projects-1234 and Name: ProjectName
    @staticmethod
    def set_radio_items_frame(parent, list_items: list, default_id: str = None, title="Select one: "):
        items_frame = tk.Frame(parent)
        tk.Label(items_frame, text=title, bd=2) \
            .grid(row=CommonWidgets.ROW_0, columnspan=CommonWidgets.ROW_SIZE, sticky=tk.W)
        id_var = tk.StringVar()
        list_items_copy = copy.deepcopy(list_items)
        list_items_copy.sort(key=lambda one_item: one_item.get(name_key).lower())
        item = find_item(lst=list_items_copy, key=id_key, value=default_id)
        if item:
            id_var.set(default_id)
        else:
            id_var.set(list_items_copy[0].get(id_key))
        for index, item in enumerate(list_items_copy):
            tk.Radiobutton(items_frame, text=f"{item.get(name_key)}", variable=id_var, value=item.get(id_key),
                           justify=tk.LEFT, command=lambda *args: None) \
                .grid(row=CommonWidgets.ROW_1 + int(index / CommonWidgets.ROW_SIZE),
                      column=index % CommonWidgets.ROW_SIZE, sticky=tk.W)  # , columnspan=1
        items_frame.grid(sticky=tk.W)
        return id_var

    @staticmethod
    def select_all_or_none(keys_list: list, keys_var_dict: dict, is_select: bool = True):
        for key in keys_list:
            keys_var_dict.get(key).set(CommonWidgets.SELECTED if is_select else CommonWidgets.UNSELECTED)

    @staticmethod
    def set_title_select_deselect_all(parent_frame, keys_list: list, keys_var_dict: dict, title: str = "Select: "):
        tk.Label(parent_frame, text=title, bd=2) \
            .grid(row=CommonWidgets.ROW_0, sticky=tk.W, columnspan=CommonWidgets.SPAN_3)
        tk.Button(
            parent_frame, text='Select all',
            command=lambda: CommonWidgets.select_all_or_none(
                keys_list=keys_list, keys_var_dict=keys_var_dict, is_select=True)) \
            .grid(row=CommonWidgets.ROW_0, column=CommonWidgets.SPAN_3, sticky=tk.W, columnspan=CommonWidgets.SPAN_1)
        tk.Button(
            parent_frame, text='Deselect all',
            command=lambda: CommonWidgets.select_all_or_none(
                keys_list=keys_list, keys_var_dict=keys_var_dict, is_select=False)) \
            .grid(row=CommonWidgets.ROW_0, column=CommonWidgets.SPAN_3 + CommonWidgets.SPAN_1, sticky=tk.W,
                  columnspan=CommonWidgets.SPAN_1)

    @staticmethod
    def set_each_check_box(parent_frame, check_box_name: str, check_box_key: str, keys_var_dict: dict, index: int,
                           default_boxes: list):
        keys_var_dict[check_box_key] = tk.StringVar()
        tk.Checkbutton(parent_frame, text=check_box_name, variable=keys_var_dict.get(check_box_key)) \
            .grid(row=int(CommonWidgets.ROW_1 + index / CommonWidgets.ROW_SIZE),
                  column=index % CommonWidgets.ROW_SIZE, sticky=tk.W)
        if check_box_key in default_boxes:
            keys_var_dict.get(check_box_key).set(CommonWidgets.SELECTED)
        else:
            keys_var_dict.get(check_box_key).set(CommonWidgets.UNSELECTED)

    @staticmethod
    def set_check_names_frame(parent, list_names: list, default_names: list = None, title="Select:"):
        if default_names is None:
            default_names = []
        names_var_dict = {}
        names_frame = tk.Frame(parent)
        list_names_copy = copy.deepcopy(list_names)
        list_names_copy.sort()
        CommonWidgets.set_title_select_deselect_all(
            names_frame, keys_list=list_names_copy, keys_var_dict=names_var_dict, title=title)
        for index, name in enumerate(list_names_copy):
            CommonWidgets.set_each_check_box(
                names_frame, check_box_name=name, check_box_key=name, keys_var_dict=names_var_dict, index=index,
                default_boxes=default_names)
        names_frame.grid(sticky=tk.W)
        return names_var_dict

    @staticmethod
    def set_check_items_frame(parent, items_list: list, default_ids: list = None, title="Select:"):
        if default_ids is None:
            default_ids = []
        ids_var_dict = {}
        items_frame = tk.Frame(parent)
        items_list_copy = copy.deepcopy(items_list)
        items_list_copy.sort(key=lambda one_item: one_item.get(name_key).lower())
        ids_list = [item.get(id_key) for item in items_list_copy]
        CommonWidgets.set_title_select_deselect_all(
            items_frame, keys_list=ids_list, keys_var_dict=ids_var_dict, title=title)
        for index, item in enumerate(items_list_copy):
            CommonWidgets.set_each_check_box(
                items_frame, check_box_name=item.get(name_key), check_box_key=item.get(id_key),
                keys_var_dict=ids_var_dict, index=index, default_boxes=default_ids)
        items_frame.grid(sticky=tk.W)
        return ids_var_dict

    @staticmethod
    def set_combobox_items_frame(parent, texts_list: list, bind_func, default_text: str = None,
                                 title: str = "Select: ", width=WIDTH_80):
        text_var = tk.StringVar()
        if not texts_list:
            return text_var
        items_frame = tk.Frame(parent)
        tk.Label(items_frame, text=title, bd=2).grid(sticky=tk.W)
        item_combobox = ttk.Combobox(items_frame, width=width, textvariable=text_var)
        texts_list_copy = copy.deepcopy(texts_list)
        item_combobox[CommonWidgets.COMBOBOX_VALUES_KEY] = tuple(texts_list_copy)
        default_index = 0
        for index, text in enumerate(texts_list_copy):
            if text == default_text:
                default_index = index
        item_combobox.current(default_index)
        item_combobox.grid(sticky=tk.W)
        item_combobox.bind("<<ComboboxSelected>>", bind_func)
        items_frame.grid(sticky=tk.W)
        return text_var

    @staticmethod
    def directional_separator(parent, title: str):
        ttk.Separator(parent, orient=tk.HORIZONTAL).grid(sticky=tk.EW)
        tk.Label(parent, text=f"\u21D3     \u21D3     \u21D3     \u21D3     \u21D3      {title}"
                              f"      \u21D3     \u21D3     \u21D3     \u21D3     \u21D3",
                 bd=2, relief="groove").grid(sticky=tk.EW)
        ttk.Separator(parent, orient=tk.HORIZONTAL).grid(sticky=tk.EW)

    @staticmethod
    def set_text_entry(parent, title: str, text_var: tk.StringVar):
        tk.Label(parent, text=title).grid(sticky=tk.W)
        text_entry = tk.Entry(parent, width=CommonWidgets.WIDTH_120, textvariable=text_var)
        text_entry.grid(sticky=tk.W)
        return text_entry



import copy
import glob
import logging
import os
import re
from time import gmtime, strftime

from octopus_python_client.common import name_key, tags_key, id_key, item_type_tag_sets, item_type_projects, Common, \
    item_type_deployment_processes, deployment_process_id_key, scope_values_key, inside_space_download_types, \
    inside_space_clone_types, version_key, item_type_library_variable_sets, item_type_variables, variable_set_id_key, \
    item_type_tenant_variables, canonical_tag_name_key, item_type_tags, tenant_id_key, item_type_migration, space_map, \
    item_type_tenants, slash_sign, underscore_sign, item_type_feeds, secret_key_key, new_value_key, hyphen_sign, \
    item_type_channels, project_id_key, item_type_releases, item_type_artifacts, file_name_key, item_type_runbooks, \
    runbook_process_id_key, item_type_accounts, token_key, space_id_key, item_type_packages, \
    item_type_scoped_user_roles, user_role_id_key, runbook_process_prefix, published_runbook_snapshot_id_key, \
    item_id_prefix_to_type_dict, positive_integer_regex, team_id_key, outer_space_clone_types, item_type_users, \
    item_type_spaces, default_password, is_service_key, space_managers_teams, item_type_teams, package_id_key, \
    comma_sign, cloned_from_project_id, item_type_runbook_processes, item_type_project_triggers, file_extension_key, \
    feed_id_key, item_types_with_logo, item_type_logo
from octopus_python_client.config import Config
from octopus_python_client.constants import Constants
from octopus_python_client.utilities.helper import find_item, save_file, find_matched_sub_list, log_raise_value_error
from octopus_python_client.utilities.send_requests_to_octopus import login_payload_user_name_key, \
    login_payload_password_key


class Migration:
    def __init__(self, src_config: Config, dst_config: Config, logger: logging.Logger = None):
        self.logger = logger if logger else logging.getLogger(self.__class__.__name__)
        self._src_config = src_config
        self._dst_config = dst_config
        self._src_common = Common(config=src_config)
        self._dst_common = Common(config=dst_config)

        self._all_types = inside_space_download_types
        self._dst_id_payload_dict = {}
        self._dst_tenant_variables_payload_dict = {}
        self._project_id = None
        self._spaces_dict = {}
        self._src_id_payload_dict = {}
        self._src_id_type_dict = {}
        self._src_id_vs_dst_id_dict = {}
        self._src_tenant_variables_payload_dict = {}
        self._type_dst_list_items_dict = {}
        self._type_full_func_dict = {}
        self._type_post_func_dict = {}
        self._type_prep_func_dict = {}
        self._type_src_list_items_dict = {}

    # search the type in the space and see if the matched item already exists
    def _find_matched_dst_item_by_src_item(self, src_item_with_dst_ids, item_type):
        self.logger.info(f"Look for the matched destination {item_type} {src_item_with_dst_ids.get(name_key)} in "
                         f"{self._dst_config.space_id}")
        dst_list_items = self._dst_common.get_list_from_one_type(item_type=item_type)
        item_name = src_item_with_dst_ids.get(name_key)
        match_dict = None
        # "channels" and "runbooks" are special, the name is not unique across a space;
        # we must use both name and project id to find the match
        if item_type == item_type_channels or item_type == item_type_runbooks or \
                item_type == item_type_project_triggers:
            match_dict = {name_key: item_name, project_id_key: src_item_with_dst_ids.get(project_id_key)}
        # type "releases" has no name and is unique by "Version" and "ProjectId"
        elif item_type == item_type_releases:
            match_dict = {version_key: src_item_with_dst_ids.get(version_key),
                          project_id_key: src_item_with_dst_ids.get(project_id_key)}
        # https://help.octopus.com/t/scopeduserrole-api-does-not-match-swagger-doc/24980
        elif item_type == item_type_scoped_user_roles:
            match_dict = {user_role_id_key: src_item_with_dst_ids.get(user_role_id_key),
                          team_id_key: src_item_with_dst_ids.get(team_id_key)}
            if src_item_with_dst_ids.get(space_id_key):
                match_dict[space_id_key] = src_item_with_dst_ids.get(space_id_key)
        # TODO type "artifacts" has no name and unique by Filename and ServerTaskId, so it is not cloneable
        elif item_type == item_type_artifacts:
            return find_item(lst=dst_list_items, key=file_name_key, value=src_item_with_dst_ids.get(file_name_key))
        elif item_type == item_type_users:
            match_dict = {login_payload_user_name_key: src_item_with_dst_ids.get(login_payload_user_name_key)}
        elif item_type == item_type_teams and src_item_with_dst_ids.get(space_id_key):
            match_dict = {name_key: item_name, space_id_key: src_item_with_dst_ids.get(space_id_key)}
        elif item_type == item_type_packages:
            match_dict = {package_id_key: src_item_with_dst_ids.get(package_id_key),
                          version_key: src_item_with_dst_ids.get(version_key)}
        elif item_name:
            match_dict = {name_key: item_name}
        else:
            log_raise_value_error(local_logger=self.logger, item=src_item_with_dst_ids,
                                  err=f"{item_type} does not have name or other keys for matching!")
        matched_sub_list = find_matched_sub_list(lst=dst_list_items, match_dict=match_dict, ignore_case=True)
        if matched_sub_list:
            if len(matched_sub_list) > 1:
                if item_type == item_type_scoped_user_roles:
                    self._dst_common.log_info_print(
                        local_logger=self.logger, item=matched_sub_list,
                        msg=f"{item_type_scoped_user_roles} could have multiple matched items in the destination space."
                            f" They are mostly just duplicate, so just return the first matched one")
                    return matched_sub_list[0]
                else:
                    log_raise_value_error(local_logger=self.logger, item=matched_sub_list,
                                          err=f"For {item_type} {item_name}, more than one item found in "
                                              f"{self._dst_config.space_id}")
            else:
                return matched_sub_list[0]
        else:
            return {}

    # For tagsets, if the destination space already have the same tagset,
    # each tag of each tagset must use the destination Tag ID not the source Tag ID,
    # so we have to replace each tag id with the destination ones,
    # otherwise it shows"Tag Id is in invalid format."
    # For additional Tag ID in source, we need to remove the ID from the source tag
    def _prepare_tag_set(self, src_item):
        src_list_tags = src_item.get(tags_key, [])
        # the source tag id should be all removed
        for src_tag in src_list_tags:
            src_tag.pop(id_key, None)
        dst_tag_set = self._find_matched_dst_item_by_src_item(src_item_with_dst_ids=src_item,
                                                              item_type=item_type_tag_sets)
        if dst_tag_set:
            dst_list_tags = dst_tag_set.get(tags_key, [])
            for src_tag in src_list_tags:
                matched_dst_tag = find_item(lst=dst_list_tags, key=name_key, value=src_tag.get(name_key))
                if matched_dst_tag:
                    src_tag[id_key] = matched_dst_tag.get(id_key)

    # we do not want to clone the child items first;
    # child items will be created automatically when parent item is created
    def _prepare_project(self, src_item):
        self.logger.info(
            f"prepare {item_type_projects} {src_item.get(name_key)} for migrating to {self._dst_config.space_id}")
        src_item.pop(deployment_process_id_key, None)
        src_item.pop(variable_set_id_key, None)
        src_item.pop(cloned_from_project_id, None)  # to avoid clone projects we do not need
        self._dst_common.prepare_project_versioning_strategy(project=src_item)

    # we do not want to clone the child items first;
    # child items will be created automatically when parent item is created
    def _prepare_library_variable_set(self, src_item):
        self.logger.info(f"prepare {item_type_library_variable_sets} {src_item.get(name_key)} for migrating to "
                         f"{self._dst_config.space_id}")
        src_item.pop(variable_set_id_key, None)

    def _replace_secrets(self, src_item):
        if secret_key_key in src_item and not src_item.get(secret_key_key).get(new_value_key):
            src_item.get(secret_key_key)[new_value_key] = default_password
            self.logger.info(f"assigned a placeholder {default_password} to {new_value_key} for {secret_key_key}")
        if token_key in src_item and not src_item.get(token_key).get(new_value_key):
            src_item.get(token_key)[new_value_key] = default_password
            self.logger.info(f"assigned a placeholder {default_password} to {new_value_key} for {token_key}")

    # SecretKey: NewValue: null; must be replaced by a placeholder string
    def _prepare_feed(self, src_item):
        self.logger.info(
            f"prepare {item_type_feeds} {src_item.get(name_key)} for migrating to {self._dst_config.space_id}")
        self._replace_secrets(src_item=src_item)

    # Token: NewValue: null; must be replaced by a placeholder string
    def _prepare_account(self, src_item):
        self.logger.info(
            f"prepare {item_type_accounts} {src_item.get(name_key)} for migrating to {self._dst_config.space_id}")
        self._replace_secrets(src_item=src_item)

    # runbook is a new type introduced in Octopus 2019.11; the older Octopus server may not support it
    # Need to upgrade Octopus server to the latest version
    def _prepare_runbook(self, src_item):
        self.logger.info(
            f"prepare {item_type_runbooks} {src_item.get(name_key)} {src_item.get(id_key)} for migrating to "
            f"{self._dst_config.space_id}: popping {runbook_process_id_key}")
        src_item.pop(runbook_process_id_key, None)
        src_item.pop(published_runbook_snapshot_id_key, None)

    def _prepare_space(self, src_item):
        # self.logger.info(f"prepare {item_type_spaces} {src_item.get(name_key)} {src_item.get(id_key)} for migrating "
        #                  f"from {self._src_config.endpoint} to {self._dst_config.endpoint} by popping "
        #                  f"{space_managers_team_members} and {space_managers_teams}")
        # src_item.pop(space_managers_team_members, None)
        # src_item.pop(space_managers_teams, None)
        self.logger.info(f"prepare {item_type_spaces} {src_item.get(name_key)} {src_item.get(id_key)} for migrating "
                         f"from {self._src_config.endpoint} to {self._dst_config.endpoint} by removing space default "
                         f"team in {space_managers_teams}")
        list_team_ids = src_item.get(space_managers_teams)
        if list_team_ids:
            for i in range(len(list_team_ids) - 1, -1, -1):
                if list_team_ids[i].endswith(src_item.get(id_key)):
                    del list_team_ids[i]

    def _prepare_user(self, src_item):
        self.logger.info(f"prepare {item_type_users} {src_item.get(login_payload_user_name_key)} {src_item.get(id_key)}"
                         f" for migrating from {self._src_config.endpoint} to {self._dst_config.space_id} by adding "
                         f"default {login_payload_password_key} as {default_password}")
        if not src_item.get(is_service_key):
            src_item[login_payload_password_key] = default_password

    def _clone_item_to_space(self, item_type, item_name=None, item_id=None, pars_dict: dict = None):
        item_badge = item_name if item_name else item_id
        self.logger.info(
            f"clone {item_type} {item_badge} from {self._src_config.space_id} to {self._dst_config.space_id} with "
            f"parameters {pars_dict}")

        full_process = self._type_full_func_dict.get(item_type)
        if full_process:
            self.logger.info(
                f"Special full processing for {item_type} {item_name} {item_id} in space {self._dst_config.space_id} "
                f"with function {full_process.__name__}")
            return full_process(item_type=item_type, item_name=item_name, item_id=item_id, pars_dict=pars_dict)

        # find the source item in file/memory
        src_item = {}
        if item_name:
            src_list_items = self._type_src_list_items_dict.get(item_type)
            src_item = find_item(lst=src_list_items, key=name_key, value=item_name)
        if not src_item:
            src_item = self._src_id_payload_dict.get(item_id)
        if not src_item:
            raise ValueError(f"{item_type} {item_badge} does not exist in the source space")

        if src_item.get(name_key) and pars_dict and pars_dict.get(Constants.NEW_ITEM_NAME_KEY):
            msg = f"clone/rename {item_type} {src_item.get(name_key)} to {pars_dict.get(Constants.NEW_ITEM_NAME_KEY)}"
            self._dst_common.log_info_print(msg=msg)
            src_item[name_key] = pars_dict.get(Constants.NEW_ITEM_NAME_KEY)

        if src_item.get(project_id_key) and pars_dict and isinstance(pars_dict.get(Constants.PROJECT_IDS_KEY), list):
            msg = f"cloning {item_type} {src_item.get(name_key)} to {pars_dict.get(Constants.PROJECT_IDS_KEY)}"
            self._dst_common.log_info_print(msg=msg)
            for project_id in pars_dict.get(Constants.PROJECT_IDS_KEY):
                self._project_id = project_id
                msg = f"cloning {item_type} {src_item.get(name_key)} to {project_id}"
                self._dst_common.log_info_print(msg=msg)
                src_item[project_id_key] = project_id
                self._src_id_vs_dst_id_dict.pop(item_id, None)
                self._create_item_to_space(item_type=item_type, src_item=src_item)
                self._project_id = None
            return

        return self._create_item_to_space(item_type=item_type, src_item=src_item)

    def _clone_single_package(self, src_package_copy_dict):
        if self._src_config.local_data:
            self._dst_common.log_info_print(
                local_logger=self.logger,
                msg=f"loading file from {self._src_config.data_path}/{self._src_config.space_id}/"
                    f"{item_type_packages}/{src_package_copy_dict.get(id_key)}")
            content = self._src_common.open_local_package(package_dict=src_package_copy_dict)
        else:
            content = self._src_common.get_package(src_package_copy_dict.get(id_key))
        file_name = Common.construct_package_name(package_dict=src_package_copy_dict)
        dst_package = self._dst_common.post_package(file_name=file_name, content=content)
        if isinstance(dst_package, dict) and dst_package.get(id_key):
            self._dst_common.log_info_print(msg=f"the destination package {file_name} was created or overwritten")
            self._src_id_vs_dst_id_dict[src_package_copy_dict.get(id_key)] = dst_package.get(id_key)
            return dst_package
        else:
            self._dst_common.log_info_print(msg=f"the destination package {file_name} existed and was skipped")
            self._src_id_vs_dst_id_dict[src_package_copy_dict.get(id_key)] = src_package_copy_dict.get(id_key)
            return src_package_copy_dict

    def _clone_package(self, src_package_copy_dict: dict, src_package_dict: dict):
        # TODO fix clone historical packages from local_data
        if self._dst_config.package_history and not self._src_config.local_data:
            self._dst_common.log_info_print(
                msg=f"cloning all versions of {item_type_packages} {src_package_dict.get(package_id_key)}...")
            package_history_list = self._src_common.get_package_history_list(package_dict=src_package_dict)
            package_history_list.reverse()
            dst_package = {}
            for package_dict in package_history_list:
                self._dst_common.log_info_print(msg=f"cloning {item_type_packages} {package_dict.get(id_key)}...")
                package_dict[file_extension_key] = src_package_copy_dict.get(file_extension_key)
                package_dict[feed_id_key] = src_package_copy_dict.get(feed_id_key)
                dst_package = self._clone_single_package(src_package_copy_dict=package_dict)
            return dst_package
        else:
            self._dst_common.log_info_print(
                msg=f"cloning the latest version of {item_type_packages} {src_package_dict.get(id_key)}...")
            src_package_copy_dict[id_key] = src_package_dict.get(id_key)
            return self._clone_single_package(src_package_copy_dict=src_package_copy_dict)

    def _put_post_item_to_space(self, item_type, src_item_copy, src_item):
        src_id_value = src_item.get(id_key)
        src_item_name = src_item.get(name_key)
        self.logger.info(
            f"check if {item_type} {src_item_name} {src_id_value} in space {self._src_config.space_id} already exists"
            f" in space {self._dst_config.space_id}")
        dst_item = self._find_matched_dst_item_by_src_item(src_item_with_dst_ids=src_item_copy, item_type=item_type)
        dst_item_exist = True if dst_item else False
        if dst_item_exist:
            if self._dst_config.overwrite:
                self._dst_common.log_info_print(
                    local_logger=self.logger,
                    msg=f"destination space {self._dst_config.space_id} already has {item_type} {src_item_name} "
                        f"{dst_item.get(id_key)}, overwriting it per user request...")
                src_item_copy[id_key] = dst_item.get(id_key)

                # TODO bug in Octopus: PUT a runbook with null RunbookProcessId will remove RunbookProcessId from dst
                if item_type == item_type_runbooks and src_item.get(runbook_process_id_key):
                    matched_runbook_process_id = runbook_process_prefix + hyphen_sign + dst_item.get(id_key)
                    self.logger.warning(f"TODO bug in Octopus: PUT a runbook with null RunbookProcessId will remove "
                                        f"RunbookProcessId from dst; Reassign process id {matched_runbook_process_id} "
                                        f"for {src_item_copy.get(id_key)}")
                    src_item_copy[runbook_process_id_key] = matched_runbook_process_id

                # # TODO for update/put the version must match (maybe need later)
                # # dst_item.get(version_key) could be zero
                # if dst_item.get(version_key) is not None:
                #     self.logger.info(f"{version_key} is updated to {dst_item.get(version_key)}")
                #     src_item_copy[version_key] = dst_item.get(version_key)

                # sometimes, overwrite may not be successful due to different reasons, we can skip in most cases
                try:
                    dst_item = self._dst_common.put_single_item(item_type=item_type, payload=src_item_copy)
                    self._dst_common.log_info_print(
                        local_logger=self.logger,
                        msg=f"{item_type} {src_item_name} {src_id_value} in space {self._src_config.space_id} "
                            f"overwrote {dst_item.get(id_key)} in {self._dst_config.space_id} successfully")
                except Exception as err:
                    self._dst_common.log_error_print(
                        local_logger=self.logger, item=src_item_copy,
                        msg=f"Failed to overwrite from {item_type} {src_item_name} {src_id_value} in space "
                            f"{self._src_config.space_id} to {dst_item.get(id_key)} in {self._dst_config.space_id} "
                            f"with {err}")
            else:
                self._dst_common.log_info_print(local_logger=self.logger,
                                                msg=f"{self._dst_config.space_id} already has {item_type} "
                                                    f"{src_item_name} {dst_item.get(id_key)}, skip it per user request")
        else:
            self.logger.info(f"destination space {self._dst_config.space_id} does not have {item_type} {src_item_name} "
                             f"{src_id_value} from space {self._src_config.space_id}, so creating it...")
            # ignore error and continue to process other items
            try:
                dst_item = self._dst_common.post_single_item(item_type=item_type, payload=src_item_copy)
                self._dst_common.log_info_print(
                    local_logger=self.logger,
                    msg=f"{item_type} {src_item_name} {src_id_value} in space {self._src_config.space_id} was cloned "
                        f"to space {self._dst_config.space_id} as {dst_item.get(id_key)} successfully")
            except Exception as err:
                self._dst_common.log_error_print(
                    local_logger=self.logger, item=src_item_copy,
                    msg=f"Failed to clone {item_type} {src_item_name} {src_id_value} from space "
                        f"{self._src_config.space_id} to space {self._dst_config.space_id} with {err}")
                return None, False
        return dst_item, dst_item_exist

    def _create_item_to_space(self, item_type, src_item):
        src_id_value = src_item.get(id_key)
        src_item_name = src_item.get(name_key)

        if src_id_value in self._src_id_vs_dst_id_dict:
            dst_id_value = self._src_id_vs_dst_id_dict.get(src_id_value)
            self._dst_common.log_info_print(
                local_logger=self.logger,
                msg=f"{item_type} {src_item_name} {src_id_value} in space {self._src_config.space_id} has already "
                    f"been cloned to space {self._dst_config.space_id} as {dst_id_value} in this session; skip it")
            return dst_id_value

        self._dst_common.log_info_print(local_logger=self.logger,
                                        msg=f"cloning {item_type} {src_item_name} {src_id_value} in space "
                                            f"{self._src_config.space_id} to space {self._dst_config.space_id}")
        self.logger.info(
            f"preprocessing {item_type} {src_item_name} {src_id_value} in space {self._src_config.space_id} with the "
            f"new references in {self._dst_config.space_id}...")
        # do not modify the items in memory
        src_item_copy = copy.deepcopy(src_item)

        # some type needs additional prep-processing
        prep_process = self._type_prep_func_dict.get(item_type)
        if prep_process:
            self.logger.info(f"special preprocessing for {item_type} {src_item_name} {src_id_value} in space "
                             f"{self._src_config.space_id} with function {prep_process.__name__}")
            prep_process(src_item=src_item_copy)

        # since the destination "Id: Self-2" has not been created and saved into map
        # we do not want to recursively replace the "Id: Self-1" of the payload;
        # it would cause infinite stack and overflow
        src_item_copy.pop(id_key, None)

        self._replace_ids(dict_list=src_item_copy)

        if item_type == item_type_packages:
            try:
                dst_item = self._clone_package(src_package_copy_dict=src_item_copy, src_package_dict=src_item)
            except Exception as err:
                self._dst_common.log_error_print(
                    local_logger=self.logger, item=src_item_copy,
                    msg=f"Failed to clone package {src_item.get(id_key)}; is history packages enabled? "
                        f"{self._dst_config.package_history}; error: {err}")
                dst_item = {}
            return dst_item.get(id_key)

        dst_item, dst_item_exist = self._put_post_item_to_space(item_type=item_type, src_item_copy=src_item_copy,
                                                                src_item=src_item)
        if not dst_item:
            return None
        dst_id_value = dst_item.get(id_key)
        self.logger.info(f"add the id pair ({src_id_value}, {dst_id_value}) to the id map")
        self._src_id_vs_dst_id_dict[src_id_value] = dst_id_value
        if item_type == item_type_spaces:
            self._spaces_dict[src_id_value] = dst_id_value
        self._dst_id_payload_dict[dst_id_value] = dst_item

        post_process = self._type_post_func_dict.get(item_type)
        if post_process and (not dst_item_exist or self._dst_config.overwrite):
            self._dst_common.log_info_print(
                local_logger=self.logger,
                msg=f"Additional post-processing for {item_type} {src_item_name} {dst_id_value} in space "
                    f"{self._dst_config.space_id} with function {post_process.__name__}")
            post_process(src_id=src_id_value, dst_id=dst_id_value)

        if item_type in item_types_with_logo and (not dst_item_exist or self._dst_config.overwrite):
            try:
                self._clone_logos(item_type=item_type, src_id=src_id_value, dst_id=dst_id_value)
            except Exception as err:
                self._dst_common.log_error_print(
                    local_logger=self.logger,
                    msg=f"Failed to clone logo for {item_type} {src_id_value} to {dst_id_value} with {err}")

        return dst_id_value

    def _clone_logos(self, item_type: str, src_id: str, dst_id: str):
        self._dst_common.log_info_print(
            local_logger=self.logger, msg=f"cloning logo for type {item_type} from {src_id} to {dst_id}")
        if self._src_config.local_data:
            pattern = self._src_common.local_logo_file(item_type=item_type, item_id=src_id, ext="*")
            local_logo_files = glob.glob(pattern)
            if not local_logo_files or len(local_logo_files) == 0:
                self._dst_common.log_info_print(
                    local_logger=self.logger, msg=f"no logo file found for {item_type} {src_id}")
                return
            self._dst_common.log_info_print(local_logger=self.logger, msg=f"loading logo file {local_logo_files[0]}")
            content = open(local_logo_files[0], 'rb')
            file_name = os.path.basename(local_logo_files[0])
        else:
            content, ext = self._src_common.get_logo(item_type=item_type, item_id=src_id)
            file_name = f"{src_id}_{item_type_logo}.{ext}"
            self._dst_common.log_info_print(local_logger=self.logger, msg=f"named in-memory logo image as {file_name}")
        dst_response = self._dst_common.post_logo(
            item_type=item_type, item_id=dst_id, file_name=file_name, content=content)
        self._dst_common.log_info_print(
            local_logger=self.logger, item=dst_response, msg=f"logo was cloned successfully")

    def _clone_type_to_space(self, item_type):
        if not item_type:
            raise ValueError("item_type must not be empty!")
        self._dst_common.log_info_print(local_logger=self.logger,
                                        msg=f"cloning items in type {item_type} from {self._src_config.space_id} to "
                                            f"{self._dst_config.space_id}")
        src_list_items = self._type_src_list_items_dict.get(item_type)
        if not src_list_items:
            self.logger.warning(f"{item_type} has no items in {self._src_config.space_id}, so skip processing it")
            return

        for src_item in src_list_items:
            self._create_item_to_space(item_type=item_type, src_item=src_item)

        self.logger.info(f"Get the updated {item_type} in {self._dst_config.space_id}")
        dst_all_items = self._dst_common.get_one_type_ignore_error(item_type=item_type)
        dst_list_items = self._dst_common.get_list_items_from_all_items(all_items=dst_all_items)
        self._type_dst_list_items_dict[item_type] = dst_list_items

    # some id like "Environments-123" should be removed if this is not referenced anyway
    # most likely the entity has been removed and the reference is broken
    # TODO it is a dangerous operation; find a better way to warn people to fix the source space first
    def _check_broken_item_id(self, string):
        # if the string is one of the src or dst item ids, it is valid
        if string in self._src_id_vs_dst_id_dict.keys() or string in self._src_id_vs_dst_id_dict.values():
            return False
        for id_prefix, item_type in item_id_prefix_to_type_dict.items():
            pattern = r"{}".format("^" + id_prefix + positive_integer_regex)
            if re.match(pattern, string):
                msg = f"***** Please remove the broken reference id {string} in source space " \
                      f"{self._src_config.space_id} *****"
                self._dst_common.log_warn_print(local_logger=self.logger, msg=msg)
                return True
        return False

    # recursively replace the old link id with the new link id
    # issue: if someone crazy names an environment name as "Environments-1", not "Development" or "Prod" etc,
    # it could cause the environment renamed as "Environment-10"
    # if id_links_map contains "Environments-1" as the key and "Environments-10" as the value
    def _replace_ids(self, dict_list):
        if isinstance(dict_list, dict):
            # self.logger.info(f"dict_list is a dict, replace ids or delete broken ids")
            # directly use dict_list.items() or dict_list.keys() could cause unexpected result
            # due to dict.pop(key) while iterating
            keys = copy.deepcopy(list(dict_list.keys()))
            for key in keys:
                value = dict_list.get(key)
                if isinstance(value, str):
                    if value in self._src_id_vs_dst_id_dict:
                        dict_list[key] = self._src_id_vs_dst_id_dict.get(value)
                        self.logger.info(
                            f"the reference value {value} has already been created in space {self._dst_config.space_id}"
                            f" as {dict_list.get(key)}")
                    elif value in self._src_id_payload_dict:
                        self.logger.info(
                            f"the reference value {value} exists in the source space but has not been cloned "
                            f"into the destination space, so clone it first in the destination space")
                        dict_list[key] = self._clone_item_to_space(item_type=self._src_id_type_dict.get(value),
                                                                   item_id=value)
                        self.logger.info(f"the reference value {value} was cloned/found and replaced with "
                                         f"{dict_list.get(key)}")
                    # remove the broken reference ids
                    elif self._check_broken_item_id(string=value):
                        self.logger.warning(
                            f"the reference value {value} is a broken reference id, so assign null to it")
                        dict_list[key] = None
                else:
                    self._replace_ids(dict_list=value)
                if isinstance(key, str):
                    if key in self._src_id_vs_dst_id_dict:
                        new_key = self._src_id_vs_dst_id_dict.get(key)
                        dict_list[new_key] = dict_list.pop(key)
                        self.logger.info(f"the reference key {key} has already been created in space "
                                         f"{self._dst_config.space_id} as {new_key}")
                    elif key in self._src_id_payload_dict:
                        self.logger.info(f"the reference key {key} exists in the source space but has not been cloned "
                                         f"into the destination space, so clone it first in the destination space")
                        new_key = self._clone_item_to_space(item_type=self._src_id_type_dict.get(key), item_id=key)
                        dict_list[new_key] = dict_list.pop(key)
                        self.logger.info(f"the reference key {key} was cloned/found and replaced with {new_key}")
                    # remove the broken reference ids
                    elif self._check_broken_item_id(string=key):
                        self.logger.warning(f"Key {value} is a broken reference id, so pop it")
                        dict_list.pop(key)
                else:
                    self._replace_ids(dict_list=key)
        elif isinstance(dict_list, list):
            # self.logger.info(f"dict_list is a list, replace ids or delete broken ids")
            # we must do reversely to avoid unexpected result on deleting by index
            for index in range(len(dict_list) - 1, -1, -1):
                element = dict_list[index]
                if isinstance(element, str):
                    if element in self._src_id_vs_dst_id_dict:
                        dict_list[index] = self._src_id_vs_dst_id_dict.get(element)
                        self.logger.info(
                            f"the reference element {element} has already been created in space "
                            f"{self._dst_config.space_id} as {dict_list[index]}")
                    elif element in self._src_id_payload_dict:
                        self.logger.info(
                            f"the reference element {element} exists in the source space but has not been "
                            f"cloned into the destination space, so clone it first in the destination space")
                        dict_list[index] = self._clone_item_to_space(item_type=self._src_id_type_dict.get(element),
                                                                     item_id=element)
                        self.logger.info(f"the reference element {element} was cloned/found and replaced with "
                                         f"{dict_list[index]}")
                    elif self._check_broken_item_id(string=element):
                        self.logger.warning(f"element {element} is a broken reference id, so delete it; index {index}")
                        del dict_list[index]
                else:
                    self._replace_ids(dict_list=element)
        # None, boolean, integer, float etc
        else:
            pass
            # self.logger.info(f"the type is {type(dict_list)} and value is {dict_list}; skip it")

    def _clone_child(self, src_parent_id, dst_parent_id, child_type, child_id_key):
        # source item
        parent_type = self._src_id_type_dict.get(src_parent_id)
        src_parent = self._src_id_payload_dict.get(src_parent_id)
        src_child_id = src_parent.get(child_id_key)
        parent_name = src_parent.get(name_key)

        # destination item
        dst_parent = self._dst_id_payload_dict.get(dst_parent_id)
        dst_child_id = dst_parent.get(child_id_key)
        self._src_id_vs_dst_id_dict[src_child_id] = dst_child_id

        self.logger.info(
            f"cloning {parent_type} {parent_name}'s {child_id_key} {src_child_id} from {self._src_config.space_id} to "
            f"{self._dst_config.space_id}...")
        src_child = self._src_id_payload_dict.get(src_child_id)

        # ignore cloning child error due to permission and other misc issues
        try:
            # TODO Octopus bug https://help.octopus.com/t/504-gateway-time-out-on-getting-all-variables/24732
            if not src_child:
                self.logger.warning(
                    f"{src_child_id} does not exist in the memory, so get it from {self._src_config.space_id}")
                src_child = self._src_common.get_single_item_by_name_or_id(item_type=child_type, item_id=src_child_id)
                self._src_id_payload_dict[src_child_id] = src_child
            dst_child = self._dst_common.get_single_item_by_name_or_id(item_type=child_type, item_id=dst_child_id)
        except Exception as err:
            self._dst_common.log_error_print(local_logger=self.logger, msg=err)
            return dst_child_id

        src_child_copy = copy.deepcopy(src_child)
        # this is for cloning runbooks, channels, or projecttriggers from one project to other projects
        if src_child_copy.get(project_id_key) and self._project_id:
            src_child_copy[project_id_key] = self._project_id

        # TODO ScopeValues seems redundant in variables; ScopeValues defines a general scope to be used for variable set
        # but ScopeValues does not scope a specific variable at all; ScopeValues in variable set payload not very useful
        # if ScopeValues is not popped, all scopes will be cloned recursively even if some of them are not used at all  
        src_child_copy.pop(scope_values_key, None)

        self._replace_ids(dict_list=src_child_copy)

        # for update/put the version must match; dst_child.get(version_key) could be zero
        if dst_child.get(version_key) is not None:
            self.logger.info(f"child {version_key} is updated to {dst_child.get(version_key)}")
            src_child_copy[version_key] = dst_child.get(version_key)

        # sometimes, overwrite may not be successful due to different reasons, we can skip in most cases
        try:
            dst_child = self._dst_common.put_single_item(item_type=child_type, payload=src_child_copy)
            self._dst_common.log_info_print(
                local_logger=self.logger,
                msg=f"Finished overwriting {parent_type} {parent_name}'s {child_type} {src_child_id} from "
                    f"{self._src_config.space_id} to {self._dst_config.space_id} as {dst_child_id} successfully")
        except Exception as err:
            self._dst_common.log_error_print(
                local_logger=self.logger, item=src_child_copy,
                msg=f"Failed to overwrite {parent_type} {parent_name}'s {child_type} {src_child_id} from "
                    f"{self._src_config.space_id} to {self._dst_config.space_id} as {dst_child_id} with {err}")

        self._dst_id_payload_dict[dst_child_id] = dst_child
        return dst_child_id

    # deployment processes are not really created as new; once a project is created, the deployment process is
    # created automatically; so we need to copy the deployment processes from the source space into the destination
    # space matching the same projects; the API call is "PUT" not "POST"
    # same for the variables
    def _post_process_project(self, src_id, dst_id):
        self._dst_common.log_info_print(
            local_logger=self.logger,
            msg=f"clone {item_type_deployment_processes} from {item_type_projects} {src_id} in "
                f"{self._src_config.space_id} to {dst_id} in {self._dst_config.space_id}")
        self._clone_child(src_parent_id=src_id, dst_parent_id=dst_id, child_type=item_type_deployment_processes,
                          child_id_key=deployment_process_id_key)
        self._dst_common.log_info_print(
            local_logger=self.logger,
            msg=f"clone {item_type_variables} from {item_type_projects} {src_id} in {self._src_config.space_id} "
                f"to {dst_id} in {self._dst_config.space_id}")
        self._clone_child(src_parent_id=src_id, dst_parent_id=dst_id, child_type=item_type_variables,
                          child_id_key=variable_set_id_key)

    def _post_process_library_variable_set(self, src_id, dst_id):
        self._dst_common.log_info_print(
            local_logger=self.logger,
            msg=f"clone {item_type_variables} from {item_type_library_variable_sets} {src_id} in "
                f"{self._src_config.space_id} to {dst_id} in {self._dst_config.space_id}")
        self._clone_child(src_parent_id=src_id, dst_parent_id=dst_id, child_type=item_type_variables,
                          child_id_key=variable_set_id_key)

    def _post_process_runbook(self, src_id, dst_id):
        self._dst_common.log_info_print(
            local_logger=self.logger,
            msg=f"clone {item_type_runbook_processes} from {item_type_runbooks} {src_id} in "
                f"{self._src_config.space_id} to {dst_id} in {self._dst_config.space_id}")
        self._clone_child(src_parent_id=src_id, dst_parent_id=dst_id, child_id_key=runbook_process_id_key,
                          child_type=item_type_runbook_processes)

    # tenant variables is special and its id is also tenant id, such as "Tenants-401"
    # so you have to use a separate map to store the tenants variables
    # also put/post tenant variables uses a different url from put variables
    def _post_process_tenant_variables(self, src_id, dst_id):
        self._dst_common.log_info_print(local_logger=self.logger,
                                        msg=f"clone {item_type_tenant_variables} from {item_type_tenants} {src_id} in "
                                            f"{self._src_config.space_id} to {dst_id} in {self._dst_config.space_id}")
        src_tenant_variables = self._src_tenant_variables_payload_dict.get(src_id)
        src_tenant_variables_copy = copy.deepcopy(src_tenant_variables)

        self._replace_ids(dict_list=src_tenant_variables_copy)

        # sometimes, overwrite may not be successful due to different reasons, we can skip in most cases
        try:
            dst_tenant_variables = \
                self._dst_common.put_post_tenant_variables(tenant_id=dst_id, tenant_variables=src_tenant_variables_copy)
            self._dst_common.log_info_print(
                local_logger=self.logger,
                msg=f"Finished overwriting {item_type_tenant_variables} from {item_type_tenants} {src_id} in "
                    f"{self._src_config.space_id} to {dst_id} in {self._dst_config.space_id} successfully")
        except Exception as err:
            dst_tenant_variables = self._dst_common.get_tenant_variables(tenant_id=dst_id)
            self._dst_common.log_error_print(
                local_logger=self.logger, item=src_tenant_variables_copy,
                msg=f"Failed to overwrite {item_type_tenant_variables} from {item_type_tenants} {src_id} in "
                    f"{self._src_config.space_id} to {dst_id} in {self._dst_config.space_id} with {err}")

        self._dst_tenant_variables_payload_dict[dst_id] = dst_tenant_variables

    # TODO how to process new item name in pars_dict
    def _full_process_tags(self, item_type, item_id, item_name=None, pars_dict: dict = None):
        self._dst_common.log_info_print(
            local_logger=self.logger,
            msg=f"special full process - clone {item_type} {item_name} {item_id} from {self._src_config.space_id} "
                f"to {self._dst_config.space_id}; TODO: tags need to honor parameters {pars_dict}")
        tag_set_name = item_id.split(slash_sign)[0]
        src_list_tag_sets = self._type_src_list_items_dict.get(item_type_tag_sets)
        src_tag_set = find_item(lst=src_list_tag_sets, key=name_key, value=tag_set_name)
        src_tag = self._src_id_payload_dict.get(item_id)
        src_tag_copy = copy.deepcopy(src_tag)
        src_tag_copy.pop(id_key, None)
        dst_tag_set = self._find_matched_dst_item_by_src_item(src_item_with_dst_ids=src_tag_set,
                                                              item_type=item_type_tag_sets)
        if dst_tag_set:
            self.logger.info(f"{item_type_tag_sets} {dst_tag_set.get(name_key)} exists in {self._dst_config.space_id}, "
                             f"so try to find {item_id}")
            if not find_item(lst=dst_tag_set.get(tags_key), key=canonical_tag_name_key, value=item_id):
                self.logger.info(f"{item_type_tag_sets} {dst_tag_set.get(name_key)} does not have {item_id} in "
                                 f"{self._dst_config.space_id}, add it")
                dst_tag_set.get(tags_key).append(src_tag_copy)
                try:
                    dst_tag_set = self._dst_common.put_single_item(item_type=item_type_tag_sets, payload=dst_tag_set)
                    self._dst_common.log_info_print(
                        local_logger=self.logger,
                        msg=f"overwrote {item_type} {item_name} {item_id} from {self._src_config.space_id} to "
                            f"{self._dst_config.space_id} successfully")
                except Exception as err:
                    self._dst_common.log_error_print(
                        local_logger=self.logger, item=dst_tag_set,
                        msg=f"Failed to overwrite {item_type} {item_name} {item_id} from {self._src_config.space_id} to"
                            f" {self._dst_config.space_id} unsuccessfully with {err}")
            else:
                self.logger.info(f"{item_type_tag_sets} {dst_tag_set.get(name_key)} already has {item_id} in "
                                 f"{self._dst_config.space_id}, skip")
        else:
            self.logger.info(
                f"{item_type_tag_sets} {dst_tag_set.get(name_key)} does not exist in {self._dst_config.space_id}, "
                f"so create it with {item_type_tags} {item_id}")
            dst_tag_set = copy.deepcopy(src_tag_set)
            dst_tag_set.pop(id_key, None)
            dst_tag_set[tags_key] = [src_tag_copy]
            # ignore error and continue to process other items
            try:
                dst_tag_set = self._dst_common.post_single_item(item_type=item_type_tag_sets, payload=dst_tag_set)
                self._dst_common.log_info_print(
                    local_logger=self.logger,
                    msg=f"cloned {item_type} {item_name} {item_id} from {self._src_config.space_id} to "
                        f"{self._dst_config.space_id} successfully")
            except Exception as err:
                self._dst_common.log_error_print(
                    local_logger=self.logger, item=dst_tag_set,
                    msg=f"Failed to clone {item_type} {item_name} {item_id} from {self._src_config.space_id} to "
                        f"{self._dst_config.space_id} with {err}")
        self._src_id_vs_dst_id_dict[item_id] = item_id
        self._dst_id_payload_dict[item_id] = find_item(lst=dst_tag_set.get(tags_key), key=canonical_tag_name_key,
                                                       value=item_id)
        return item_id

    def _load_types(self):
        if self._src_config.local_data:
            self._dst_common.log_info_print(
                local_logger=self.logger,
                msg=f"Reading files {self._all_types} from local source {self._src_config.space_id}...")
        else:
            self._dst_common.log_info_print(
                local_logger=self.logger,
                msg=f"Downloading {self._all_types} from space {self._src_config.space_id}...")
        actual_src_space_id = None
        for item_type in self._all_types:
            if self._src_config.local_data:
                self._dst_common.log_info_print(
                    local_logger=self.logger,
                    msg=f"Loading {item_type} in {self._src_config.space_id} from the local file...")
                src_list_items = self._src_common.get_list_items_from_file(item_type=item_type)
            else:
                self._dst_common.log_info_print(
                    local_logger=self.logger,
                    msg=f"Loading {item_type} from source space {self._src_config.space_id}...")
                src_list_items = self._src_common.get_list_from_one_type(item_type=item_type)
            selected_src_list_items = []
            for src_item in src_list_items:
                if not actual_src_space_id and self._src_config.local_data and item_type == item_type_projects \
                        and src_item.get(space_id_key):
                    actual_src_space_id = src_item.get(space_id_key)
                    self.logger.info(f"the actual source space id is {actual_src_space_id}; the local one is "
                                     f"{self._src_config.space_id}")
                if not self._src_config.space_id and src_item.get(space_id_key) \
                        or item_type == item_type_scoped_user_roles \
                        and self._src_config.space_id != src_item.get(space_id_key):
                    self.logger.info(f"skip loading {item_type} - source space id {self._src_config.space_id} - source "
                                     f"item space id {src_item.get(space_id_key)}")
                    continue
                # TODO some items are a pure list of strings, they might be useful in the future, like variables/names
                if isinstance(src_item, str):
                    self.logger.warning(f"{item_type} {src_item} is ignored when loading")
                elif isinstance(src_item, dict) and src_item.get(id_key):
                    self._src_id_payload_dict[src_item.get(id_key)] = src_item
                    self._src_id_type_dict[src_item.get(id_key)] = item_type
                    selected_src_list_items.append(src_item)
                elif isinstance(src_item, dict) and src_item.get(tenant_id_key):
                    self._src_tenant_variables_payload_dict[src_item.get(tenant_id_key)] = src_item
                else:
                    log_raise_value_error(local_logger=self.logger, item=src_item, err=f"{item_type} is not valid")
            self._type_src_list_items_dict[item_type] = selected_src_list_items
        if self._src_config.local_data and not actual_src_space_id:
            log_raise_value_error(local_logger=self.logger,
                                  err=f"Could not find an actual space id inside the local source "
                                      f"{self._src_config.space_id}")
        return actual_src_space_id if actual_src_space_id else self._src_config.space_id

    def _prep_tag_sets(self):
        self._dst_common.log_info_print(local_logger=self.logger, msg="prepare for all tags")
        src_list_tag_sets = self._type_src_list_items_dict.get(item_type_tag_sets)
        for src_tag_set in src_list_tag_sets:
            for src_tag in src_tag_set.get(tags_key):
                canonical_tag_name = src_tag.get(canonical_tag_name_key)
                self._src_id_payload_dict[canonical_tag_name] = src_tag
                self._src_id_type_dict[canonical_tag_name] = item_type_tags

    def _initialize_maps(self):
        self._dst_common.log_info_print(local_logger=self.logger,
                                        msg=f"initialize the map")
        self._type_prep_func_dict[item_type_accounts] = self._prepare_account
        self._type_prep_func_dict[item_type_feeds] = self._prepare_feed
        self._type_prep_func_dict[item_type_library_variable_sets] = self._prepare_library_variable_set
        self._type_prep_func_dict[item_type_projects] = self._prepare_project
        self._type_prep_func_dict[item_type_runbooks] = self._prepare_runbook
        self._type_prep_func_dict[item_type_spaces] = self._prepare_space
        self._type_prep_func_dict[item_type_tag_sets] = self._prepare_tag_set
        self._type_prep_func_dict[item_type_users] = self._prepare_user

        self._type_post_func_dict[item_type_library_variable_sets] = self._post_process_library_variable_set
        self._type_post_func_dict[item_type_projects] = self._post_process_project
        self._type_post_func_dict[item_type_runbooks] = self._post_process_runbook
        self._type_post_func_dict[item_type_tenants] = self._post_process_tenant_variables

        self._type_full_func_dict[item_type_tags] = self._full_process_tags

        actual_src_space_id = self._load_types()

        if self._src_config.local_data:
            self._src_id_vs_dst_id_dict[actual_src_space_id] = self._dst_config.space_id
        else:
            self._src_id_vs_dst_id_dict[self._src_config.space_id] = self._dst_config.space_id

        if item_type_tag_sets in self._all_types:
            self._prep_tag_sets()

    def _save_space_map(self):
        current_time = strftime("%Y-%m-%d-%H-%M-%S", gmtime())
        local_file = self._dst_common.get_local_single_item_file(item_name=space_map + underscore_sign + current_time,
                                                                 item_type=item_type_migration)
        self._dst_common.log_info_print(local_logger=self.logger, msg=f"writing item id map to {local_file}")
        save_file(file_path_name=local_file, content=self._src_id_vs_dst_id_dict)
        self._dst_common.log_info_print(local_logger=self.logger, msg=f"***** data migration/clone is DONE! *****")

    def clone_space(self, item_types_comma_delimited=None):
        if item_types_comma_delimited:
            process_types = item_types_comma_delimited.split(comma_sign)
        else:
            process_types = inside_space_clone_types
        self._dst_common.log_info_print(
            local_logger=self.logger,
            msg=f"cloning types {process_types} from {self._src_config.space_id} on server {self._src_config.endpoint}"
                f" to {self._dst_config.space_id} on server {self._dst_config.endpoint}")
        if not self._dst_config.overwrite:
            self._dst_config.overwrite = input(
                f"Some entities may already exist in {self._dst_config.space_id} on server {self._dst_config.endpoint};"
                f" Do you want to overwrite the existing entities? "
                f"If no, we will skip the existing entities. [Y/n]: ") == 'Y'
        self._dst_config.types = process_types
        self.clone_space_types()

    def clone_space_types(self):
        self._dst_common.log_info_print(
            local_logger=self.logger,
            msg=f"cloning types {self._dst_config.types} from {self._src_config.space_id} on server "
                f"{self._src_config.endpoint} to {self._dst_config.space_id} on server {self._dst_config.endpoint}")
        self._all_types = inside_space_download_types
        self._initialize_maps()
        for item_type in self._dst_config.types:
            if item_type in inside_space_clone_types:
                self._clone_type_to_space(item_type=item_type)
        self._save_space_map()

    def clone_space_item(self, item_type, item_name=None, item_id=None):
        item_badge = item_name if item_name else item_id
        self._dst_common.log_info_print(
            local_logger=self.logger,
            msg=f"cloning {item_type} {item_badge} from {self._src_config.space_id} on server "
                f"{self._src_config.endpoint} to {self._dst_config.space_id} on server {self._dst_config.endpoint}")
        if not self._dst_config.overwrite:
            self._dst_config.overwrite = input(
                f"Some entities may already exist in {self._dst_config.space_id} on server {self._dst_config.endpoint};"
                f" Do you want to overwrite the existing entities? "
                f"If no, we will skip the existing entities. [Y/n]: ") == 'Y'
        self._dst_config.type = item_type
        self._src_config.item_id = item_id
        self._src_config.item_name = item_name
        self.clone_space_item_new_name()

    def clone_space_item_new_name(self, pars_dict: dict = None):
        item_type = self._dst_config.type
        item_name = self._src_config.item_name
        item_id = self._src_config.item_id
        item_badge = item_name if item_name else item_id
        self._dst_common.log_info_print(
            local_logger=self.logger,
            msg=f"cloning {item_type} {item_badge} from {self._src_config.space_id} on server "
                f"{self._src_config.endpoint} to {self._dst_config.space_id} on server {self._dst_config.endpoint}"
                f"with parameters {pars_dict}")
        self._all_types = inside_space_download_types
        self._initialize_maps()
        if item_type in inside_space_clone_types:
            self._clone_item_to_space(item_type=item_type, item_name=item_name, item_id=item_id, pars_dict=pars_dict)
        self._save_space_map()

    def clone_server(self, space_id_or_name_comma_delimited=None, item_types_comma_delimited=None):
        list_space_ids = self._src_common.get_list_spaces_ids_sorted(
            space_id_or_name_comma_delimited=space_id_or_name_comma_delimited)
        self._dst_common.log_info_print(
            local_logger=self.logger,
            msg=f"cloning outer space and {list_space_ids} for item types {item_types_comma_delimited} from server "
                f"{self._src_config.endpoint} to server {self._dst_config.endpoint}")

        if input(f"Are you sure you want to clone server to server? If yes, your permission on the destination "
                 f"server could be overwritten and revoked (default password is {default_password} [Y/n]: ") != 'Y':
            return

        if not self._dst_config.overwrite:
            self._dst_config.overwrite = input(
                f"Some entities may already exist in server {self._dst_config.endpoint}; "
                f"Do you want to overwrite the existing entities? "
                f"If no, we will skip the existing entities. [Y/n]: ") == 'Y'

        # outer space
        self._all_types = outer_space_clone_types
        self._initialize_maps()
        for item_type in outer_space_clone_types:
            self._clone_type_to_space(item_type=item_type)
        self._dst_common.log_info_print(item=self._spaces_dict,
                                        msg=f"see log for space id map from source server to dest server")

        # inside space
        for src_space_id in list_space_ids:
            dst_space_id = self._spaces_dict.get(src_space_id)
            self._dst_common.log_info_print(msg=f"clone {src_space_id} on server {self._src_config.endpoint} to "
                                                f"{dst_space_id} on server {self._dst_config.endpoint}")
            self._src_config.space_id = src_space_id
            self._dst_config.space_id = dst_space_id
            self.clone_space(item_types_comma_delimited=item_types_comma_delimited)
        self._save_space_map()



import tkinter as tk

from octopus_python_client.actions import ACTIONS_DICT
from octopus_python_client.common import Common


class ActionsWidgets(tk.Frame):
    def __init__(self, parent: tk.Frame, server: Common, source: Common, next_button: tk.Button = None,
                 submit_button: tk.Button = None):
        super().__init__(parent)
        self.server = server
        self.source = source

        self.next_button = next_button
        self.submit_button = submit_button

        self.select_action_variable = None

        self.update_step()

    def update_step(self):
        tk.Label(self, text="Action selection", bd=2, relief="groove").pack(side="top", fill="x")
        self.select_action_variable = tk.StringVar()
        self.select_action_variable.set(self.server.config.action)
        for action, description in ACTIONS_DICT.items():
            radio = tk.Radiobutton(self, text=f"{action} ({description})", variable=self.select_action_variable,
                                   value=action, justify=tk.LEFT, command=self.select_action)
            radio.pack(anchor=tk.W)

    def select_action(self):
        self.server.config.action = self.select_action_variable.get()

    def process_config(self):
        self.server.config.action = self.select_action_variable.get()
        self.server.config.save_config()
        return True



import logging

import requests
import urllib3

from octopus_python_client.config import Config
from octopus_python_client.utilities.helper import log_raise_value_error

operation_delete = "delete"
operation_get = "get"
operation_post = "post"
operation_put = "put"
operation_get_file = "get_file"
operation_post_file = "post_file"

application_json = "application/json"
content_type_key = "Content-Type"
login_payload_password_key = "Password"
login_payload_user_name_key = "Username"
users_login_url_suffix = "users/login"
x_octopus_api_key_key = "X-Octopus-ApiKey"

logger = logging.getLogger(__name__)


def call_octopus(config: Config, url_suffix=None, operation=None, payload=None, files=None):
    urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)
    url = config.endpoint + url_suffix if url_suffix else config.endpoint
    operation = operation if operation else operation_get
    with requests.Session() as session:
        headers = {} if operation == operation_post_file else {content_type_key: application_json}
        if config.api_key:
            headers[x_octopus_api_key_key] = config.api_key
        elif config.user_name and config.password:
            login_url = config.endpoint + users_login_url_suffix
            login_payload = {login_payload_user_name_key: config.user_name, login_payload_password_key: config.password}
            session.post(login_url, json=login_payload, headers=headers, verify=config.pem)
        else:
            log_raise_value_error(local_logger=logger, err=f"either api_key or user_name and password are required")
        try:
            logger.info(f"{operation}: " + url)
            if operation.lower() == operation_post:
                session_response = session.post(url, json=payload, headers=headers, verify=config.pem)
            elif operation.lower() == operation_get:
                session_response = session.get(url, params=payload, headers=headers, verify=config.pem)
            elif operation.lower() == operation_put:
                session_response = session.put(url, json=payload, headers=headers, verify=config.pem)
            elif operation.lower() == operation_delete:
                session_response = session.delete(url, headers=headers, verify=config.pem)
            elif operation.lower() == operation_get_file:
                session_response = session.get(url, headers=headers, verify=config.pem)
                if session_response.status_code == 200:
                    logger.info("response headers: " + str(session_response.headers))
                    return session_response.content, session_response.headers
            elif operation.lower() == operation_post_file:
                session_response = session.post(url, files=files, headers=headers, verify=config.pem)
            else:
                log_raise_value_error(local_logger=logger,
                                      err=f"Invalid operation: {operation}; only post, get, put, delete, get_file, "
                                          f"post_file are supported")
            logger.info("response status code: " + str(session_response.status_code))
            logger.info("response headers: " + str(session_response.headers))
            response_json = ""
            if session_response.text:
                response_json = session_response.json()
            if session_response.status_code < 200 or session_response.status_code > 299:
                log_raise_value_error(local_logger=logger,
                                      err=f"Error code: {session_response.status_code}; Reason: "
                                          f"{session_response.reason}; {response_json}")
            return response_json, session_response.headers
        except requests.exceptions.RequestException as e:
            log_raise_value_error(local_logger=logger, err=e)


# TODO for testing purpose, to be removed
def run():
    print("==================== test octopus http requests ===================")
    # octopus_config = {"endpoint": "https://demo.octopusdeploy.com/api/", "api_key": None, "user_name": "guest",
    #                   "password": "guest", "pem": False}
    response, headers = call_octopus(config=Config(), url_suffix="Spaces-1/environments")
    print(response)
    package_name_version = "SampleFunction.1.0.1"
    response, headers = \
        call_octopus(config=Config(), url_suffix=f"Spaces-1/packages/packages-{package_name_version}")
    package_dict = response
    print()
    print(package_dict)
    file_name = f"{package_dict.get('PackageId')}.{package_dict.get('Version')}{package_dict.get('FileExtension')}"
    print(file_name)
    content, headers = \
        call_octopus(config=Config(),
                     url_suffix=f"Spaces-1/packages/packages-{package_name_version}/raw",
                     operation=operation_get_file)
    print(len(content))

    dst_octopus_config = Config()
    dst_octopus_config.endpoint = "http://localhost/api/"
    dst_octopus_config.api_key = "API-XXX"

    upload_response, headers = \
        call_octopus(config=dst_octopus_config,
                     url_suffix="Spaces-1/packages/raw?overwriteMode=OverwriteExisting",
                     # IgnoreIfExists
                     operation=operation_post_file,
                     files={"file": (file_name, content)})
    print(upload_response)

    file = open(file_name, "wb")
    file.write(content)
    file.close()
    upload_response = call_octopus(config=Config(),
                                   url_suffix="Spaces-1/packages/raw?overwriteMode=IgnoreIfExists",
                                   operation=operation_post_file,
                                   files={"file": (file_name, open(file_name, 'rb'))})
    print(upload_response)


if __name__ == "__main__":
    run()



import json
import logging
from pprint import pformat

from octopus_python_client.common import item_type_deployment_processes, item_type_projects, id_key, \
    deployment_process_id_key, item_type_channels, packages_key, action_name_key, package_reference_name_key, Common, \
    feed_id_key, package_id_key, item_type_feeds, item_type_packages, version_key, items_key, name_key, timestamp_key, \
    value_key, project_id_key, next_version_increment_key, release_notes_key, channel_id_key, selected_packages_key, \
    item_type_releases, item_type_deployments, item_type_tenants, item_type_environments, tenant_id_key, newline_sign, \
    environment_id_key, release_id_key, comments_key, release_versions_key, url_prefix_key, dot_sign, sha_key, \
    author_key, latest_commit_sha_key, title_key
from octopus_python_client.config import Config
from octopus_python_client.utilities.helper import replace_list_new_value, parse_string, find_item
from octopus_python_client.utilities.send_requests_to_octopus import operation_post

logger = logging.getLogger(__name__)


class ReleaseDeployment:
    def __init__(self, config: Config, project_name, channel_name=None, notes=None):
        self._config = config
        self._common = Common(config=self._config)

        assert project_name, "project name must not be empty!"
        project = self._common.get_single_item_by_name(item_type=item_type_projects, item_name=project_name)
        assert project, f"Project {project_name} cannot be found or you do not have permission to access!"
        self._project_id = project.get(id_key)
        self._deployment_process_id = project.get(deployment_process_id_key)
        self._notes = notes
        self._release_request_payload = {project_id_key: self._project_id, release_notes_key: notes}

        self._channel_id = ""
        if channel_name:
            channel = self._common.find_sub_by_item(item_type=item_type_projects, item_id=self._project_id,
                                                    sub_type=item_type_channels, sub_name=channel_name)
            assert channel, f"Cannot find channel {channel_name} in project {project_name}"
            self._channel_id = channel.get(id_key)
        if self._channel_id:
            self._release_request_payload[channel_id_key] = self._channel_id

        # package versions read from a variable set,
        # e.g. release_versions = {"Name": "package.near", "Value": "20.0225.1714"}
        self._packages_variable_set_name = None
        # user selected package versions, e.g. "{'packages': {'package.near': '20.0225.1714'}}"
        self._package_version_dict = None

        self._template = None
        self._selected_packages = None
        self._release_response = None
        self._release_id = None
        self._commits_variable_set_name = "configuration_commits" + dot_sign + project_name
        self._gitlab_url_prefix = self._get_url_prefix(set_name="gitlab_info")
        self._latest_commit_dict = None

    def _get_url_prefix(self, set_name):
        info_service_list_variables = self._common.get_list_variables_by_set_name_or_id(set_name=set_name)
        if info_service_list_variables:
            url_prefix_variable = find_item(lst=info_service_list_variables, key=name_key, value=url_prefix_key)
            if url_prefix_variable:
                return url_prefix_variable.get(value_key)
        return ""

    def _get_deployment_process_template(self):
        logger.info(f"Fetching deployment process template from {self._deployment_process_id} with channel "
                    f"{self._channel_id} in {self._config.space_id}, which is used for defining release and deployment")
        address = f"{item_type_deployment_processes}/{self._deployment_process_id}/template?channel=" \
                  f"{self._channel_id}"
        self._template = self._common.request_octopus_item(address=address)

    def _get_selected_packages(self):
        logger.info(f"getting package information for each step")
        self._selected_packages = []
        for package in self._template.get(packages_key):
            logger.info(f"getting package information for {package.get(action_name_key)} with package "
                        f"{package.get(package_reference_name_key)}")
            address = f"{item_type_feeds}/{package.get(feed_id_key)}/{item_type_packages}/versions?packageId=" \
                      f"{package.get(package_id_key)}&take=1"
            package_detail = self._common.request_octopus_item(address=address)
            selected_package = {action_name_key: package.get(action_name_key),
                                package_reference_name_key: package.get(package_reference_name_key),
                                version_key: package_detail.get(items_key)[0].get(version_key)}
            self._selected_packages.append(selected_package)

    def _update_package_version(self, package, version):
        match_dict = {package_reference_name_key: package}
        replace_dict = {version_key: version}
        replace_list_new_value(lst=self._selected_packages, match_dict=match_dict, replace_dict=replace_dict)

    def _update_selected_packages(self):
        logger.info("update package versions...")
        if self._packages_variable_set_name:
            list_release_versions = self._common.get_list_variables_by_set_name_or_id(
                set_name=self._packages_variable_set_name)
            for package_version in list_release_versions:
                self._update_package_version(package=package_version.get(name_key),
                                             version=package_version.get(value_key))
        if self._package_version_dict:
            for package, version in self._package_version_dict.items():
                self._update_package_version(package=package, version=version)

    def _form_single_commit_note(self, commit_variable):
        date_time = commit_variable.get(name_key)
        commit_json = commit_variable.get(value_key)
        commit_dict = json.loads(commit_json)
        title = ". ".join(commit_dict.get(title_key)) if isinstance(commit_dict.get(title_key), list) else \
            str(commit_dict.get(title_key))
        return f"- {date_time} - [{title}]({self._gitlab_url_prefix}" \
               f"{commit_dict.get(sha_key)}) - {commit_dict.get(author_key)}"

    def _get_prev_release_match_commit_date_time(self, list_releases: list):
        if list_releases:
            for release in list_releases:
                logger.info(f"checking {release.get(id_key)} for project {self._project_id}...")
                if release.get(release_notes_key):
                    logger.info(f"found notes in {release.get(id_key)} and try to get the commit timestamp...")
                    notes_last_line = release.get(release_notes_key).splitlines()[-1]
                    last_line_parsed = parse_string(local_logger=logger, string=notes_last_line)
                    if isinstance(last_line_parsed, dict) and last_line_parsed.get(self._commits_variable_set_name):
                        prev_release_match_commit_date_time = last_line_parsed.get(self._commits_variable_set_name)
                        topic_note = f"\nThe previous release with the commit timestamp " \
                                     f"{prev_release_match_commit_date_time} is {release.get(id_key)} " \
                                     f"(release version: {release.get(version_key)}). "
                        logger.info(topic_note)
                        return topic_note, prev_release_match_commit_date_time
                    else:
                        logger.warning(f"the commit timestamp in {release.get(id_key)} not exist")
                else:
                    logger.warning(f"{release.get(id_key)} has no release notes")
            topic_note = f"\nNo previous release with the commit timestamp in notes was found"
            logger.warning(topic_note)
            return topic_note, ""
        else:
            topic_note = f"\nThis is the first release for project {self._project_id}. "
            logger.info(topic_note)
            return topic_note, ""

    def _generate_commits_notes(self):
        logger.info("generating the release notes for the commits history")
        # find the latest/previous release for this project
        list_releases = self._common.get_project_releases_sorted_list(project_id=self._project_id)
        topic_note, prev_release_match_commit_date_time = \
            self._get_prev_release_match_commit_date_time(list_releases=list_releases)
        list_notes = ["\n========== below is auto-generated notes ==========", topic_note]

        # historical commits since the latest release
        list_configuration_commits = self._common.get_list_variables_by_set_name_or_id(
            set_name=self._commits_variable_set_name)
        if not list_configuration_commits:
            msg = f"\nVariable set {self._commits_variable_set_name} contains NONE historical commits. No commits " \
                  f"can be matched to the releases."
            logger.error(msg)
            list_notes.append(msg)
            return newline_sign.join(list_notes)

        list_configuration_commits_sorted = sorted(list_configuration_commits, key=lambda k: k.get(name_key))
        list_commit_notes = []
        latest_commit_variable = None
        for commit_variable in list_configuration_commits_sorted:
            latest_commit_variable = commit_variable
            commit_note = self._form_single_commit_note(commit_variable=commit_variable)
            list_commit_notes.append(commit_note)
            # if prev release has no matched commit or the commit could not be found, append all commits
            # once the prev release matched commit is found, only append the commits after it
            if prev_release_match_commit_date_time == commit_variable.get(name_key):
                logger.info(f"found a matched timestamp {prev_release_match_commit_date_time} in commits history, "
                            f"so will start to record all the commits after it")
                list_commit_notes = []
        list_notes.append("\nThe gitlab commits since the previous release are: ")
        if not list_commit_notes:
            list_notes.append("None")
        else:
            list_commit_notes.reverse()
            list_notes.extend(list_commit_notes)

        # matched latest commit for the current release
        if latest_commit_variable:
            latest_timestamp = latest_commit_variable.get(name_key)
            self._latest_commit_dict = json.loads(latest_commit_variable.get(value_key))
            self._latest_commit_dict[timestamp_key] = latest_timestamp
            latest_commit_note = self._form_single_commit_note(commit_variable=latest_commit_variable)
            list_notes.append(f"\nThe matched latest gitlab commit for this release is {latest_commit_note}")
            list_notes.append(f"\nBelow is a python dictionary read by Octopus python client in the succeeding "
                              f"releases to identify the gitlab commit for the preceding release and it must be the "
                              f"last line in the release notes. '{self._commits_variable_set_name}' is the variable "
                              f"set name for the commits history and the value is the matched commit timestamp for "
                              f"this release")
            list_notes.append("\n{'" + f"{self._commits_variable_set_name}" + "': '" + f"{latest_timestamp}" + "'}")
        return newline_sign.join(list_notes)

    def _process_notes(self):
        logger.info("process notes...")
        notes = parse_string(local_logger=logger, string=self._notes)
        if isinstance(notes, dict):
            logger.info("the notes is a dictionary, so further process...")
            logger.info(pformat(notes))
            self._packages_variable_set_name = notes.get(release_versions_key)
            self._package_version_dict = notes.get(item_type_packages)
            self._update_selected_packages()
        commit_notes = self._generate_commits_notes()
        if self._notes:
            self._release_request_payload[release_notes_key] = newline_sign.join([self._notes, commit_notes])
        else:
            self._release_request_payload[release_notes_key] = commit_notes

    def _process_package_versions_notes(self):
        self._get_deployment_process_template()
        self._get_selected_packages()
        self._process_notes()

    # release version must be unique for each release
    def create_release(self, release_version=None):
        self._process_package_versions_notes()
        if not release_version:
            release_version = self._template.get(next_version_increment_key)
        self._release_request_payload[version_key] = release_version
        self._release_request_payload[selected_packages_key] = self._selected_packages
        logger.info("the request release payload is")
        logger.info(pformat(self._release_request_payload))
        self._release_response = self._common.request_octopus_item(address=item_type_releases,
                                                                   payload=self._release_request_payload,
                                                                   operation=operation_post)
        if self._latest_commit_dict:
            self._release_response[latest_commit_sha_key] = self._latest_commit_dict.get(sha_key)
        logger.info("the response release payload is")
        logger.info(pformat(self._release_response))
        self._common.save_single_item(item_type=item_type_releases, item=self._release_response)
        self._release_id = self._release_response.get(id_key)
        return self._release_response

    @staticmethod
    def create_deployment_direct(config: Config, environment_name, tenant_name, release_id=None, project_name=None,
                                 comments=None):
        logger.info(f"creating a deployment for {release_id} in space {config.space_id} with environment "
                    f"{environment_name}, tenant {tenant_name} and comments: {comments}")
        common = Common(config=config)

        # TODO project_name
        assert (release_id or project_name), "either release_id or project_name must exist!"
        assert environment_name, "environment_name must not be empty!"
        assert tenant_name, "tenant_name must not be empty!"

        if not release_id:
            logger.info(f"Get the latest release id for project {project_name}")
            project_id = common.get_item_id_by_name(item_type=item_type_projects, item_name=project_name)
            releases_list = common.get_project_releases_sorted_list(project_id=project_id)
            if not releases_list:
                raise ValueError(f"Project {project_name} does not have any releases. Please create a release first")
            release_id = releases_list[0].get(id_key)
            logger.info(f"The latest release id is {release_id}")

        deployment_request_payload = \
            {release_id_key: release_id,
             environment_id_key: common.get_item_id_by_name(item_type=item_type_environments,
                                                            item_name=environment_name),
             tenant_id_key: common.get_item_id_by_name(item_type=item_type_tenants, item_name=tenant_name),
             comments_key: comments}
        logger.info("the request deployment payload is")
        logger.info(pformat(deployment_request_payload))
        deployment_response_payload = common.request_octopus_item(address=item_type_deployments,
                                                                  payload=deployment_request_payload,
                                                                  operation=operation_post)
        logger.info("the response deployment payload is")
        common.log_info_print(local_logger=logger, msg=json.dumps(deployment_response_payload))
        common.save_single_item(item_type=item_type_deployments, item=deployment_response_payload)
        return deployment_response_payload

    def create_deployment_for_current_release(self, config, environment_name=None, tenant_name=None, comments=None):
        return ReleaseDeployment.create_deployment_direct(config=config, release_id=self._release_id,
                                                          environment_name=environment_name, tenant_name=tenant_name,
                                                          comments=comments)

    @property
    def release_id(self):
        logger.info(f"get the release id {self._release_id}")
        return self._release_id

    @property
    def release_response(self):
        logger.info(f"get the release response for {self._release_id}")
        return self._release_response

    def _extract_package_versions(self):
        package_versions_dict = {}
        for package in self._selected_packages:
            if package.get(package_reference_name_key):
                package_versions_dict[package.get(package_reference_name_key)] = package.get(version_key)
        return package_versions_dict

    @staticmethod
    def get_package_versions(config: Config, project_name, channel_name=None, notes=None):
        release = ReleaseDeployment(config=config, project_name=project_name, channel_name=channel_name, notes=notes)
        release._process_package_versions_notes()
        return release._extract_package_versions()

    @staticmethod
    def create_release_direct(config: Config, release_version, project_name, channel_name=None, notes=None):
        common = Common(config=config)
        release = ReleaseDeployment(config=config, project_name=project_name, channel_name=channel_name, notes=notes)
        release.create_release(release_version=release_version)
        common.log_info_print(local_logger=logger, msg=json.dumps(release.release_response))
        return release

    @staticmethod
    def create_release_deployment(config: Config, release_version, project_name, comments, channel_name=None,
                                  notes=None, environment_name=None, tenant_name=None):
        release = ReleaseDeployment.create_release_direct(
            config=config, release_version=release_version, project_name=project_name, channel_name=channel_name,
            notes=notes)
        return release.create_deployment_for_current_release(config=config, environment_name=environment_name,
                                                             tenant_name=tenant_name, comments=comments)



import argparse
import logging
import os
import sys

from octopus_python_client.actions import Actions, MIGRATION_LIST
from octopus_python_client.common import item_type_deployment_processes, outer_space_download_types, steps_key, \
    inside_space_download_types, deployment_process_id_key, Common
from octopus_python_client.config import Config, SystemConfig
from octopus_python_client.deployment_processes import DeploymentProcesses
from octopus_python_client.gui.main_gui import MainGUI
from octopus_python_client.migration import Migration
from octopus_python_client.projects import Projects
from octopus_python_client.release_deployment import ReleaseDeployment
from octopus_python_client.utilities.helper import log_raise_value_error

logger = logging.getLogger(__name__)


class OctopusClient:

    def __init__(self):
        self._target_config = Config()
        self._target_common = Common(config=self._target_config)
        self._source_config = Config(is_source_server=True)
        self._source_common = Common(config=self._source_config)

    @staticmethod
    def _parse_args():
        parser = argparse.ArgumentParser()
        parser.add_argument("-v", "--version", help="if present, print the version information", action="store_true")
        parser.add_argument("-o", "--endpoint", help="octopus endpoint")
        parser.add_argument("-s", "--space_id_name", help="octopus space id or name")
        parser.add_argument("-m", "--pem", help="octopus endpoint root pem file path; -m=false to disable pem")
        parser.add_argument("-sps", "--spaces",
                            help='list of octopus space id or name, like "my space,Spaces-1,Spaces-2"')
        parser.add_argument("-d", "--data_path",
                            help="the local path for the Octopus server data, 'current' = the current work path")
        parser.add_argument("-k", "--api_key",
                            help="api key for octopus; either api_key or user_name and password are required")
        parser.add_argument("-user", "--user_name",
                            help="user_name for octopus; either api_key or user_name and password are required")
        parser.add_argument("-pass", "--password",
                            help="password for octopus; either api_key or user_name and password are required")
        parser.add_argument("-sre", "--source_endpoint", help="source octopus endpoint for clone")
        parser.add_argument("-srd", "--source_data_path",
                            help="the local path for the source Octopus server data, 'current' = the current work path")
        parser.add_argument("-srk", "--source_api_key",
                            help="api key for octopus; either api_key or user_name and password are required")
        parser.add_argument("-sru", "--source_user_name",
                            help="user_name for octopus; either api_key or user_name and password are required")
        parser.add_argument("-srp", "--source_password",
                            help="password for octopus; either api_key or user_name and password are required")
        parser.add_argument("-srs", "--source_space_id_name",
                            help="source octopus space id or name for clone/migration")
        parser.add_argument("-srm", "--source_pem",
                            help="source octopus endpoint root pem file path; -srm=false to disable pem")
        parser.add_argument("-ld", "--local_data", help="if present, local_data = True; the source server/space "
                                                        "data are stored as YAML files locally", action="store_true")
        parser.add_argument("-a", "--action", help=str(Actions.__dict__.values()))
        parser.add_argument("-ow", "--overwrite", help="if present, overwrite = True", action="store_true")
        parser.add_argument("-ns", "--no_stdout", help="if present, no_stdout = True, means no stdout",
                            action="store_true")
        parser.add_argument("-ts", "--item_types",
                            help="if not item_types and not octopus_space_id, get all item types "
                                 "regardless whether they are above Spaces; if (not item_types) and octopus_space_id, "
                                 "get all item types below octopus_space_id; "
                                 'list like "accounts,actiontemplates,artifacts" is also accepted; '
                                 "item types above Spaces: " + ", ".join(outer_space_download_types) +
                                 "; \nitem types above and under Spaces: " + ", ".join(inside_space_download_types))
        parser.add_argument("-tp", "--item_type", help="one of item types above Spaces: " + ", ".join(
            outer_space_download_types) + "; \nitem types above and under Spaces: " + ", ".join(
            inside_space_download_types))
        parser.add_argument("-nm", "--item_name", help="item name: e.g. project_name")
        parser.add_argument("-id", "--item_id", help="item id: e.g. Lifecycles-1")
        parser.add_argument("-bn", "--base_item_name", help="base item name, either local or on Octopus server")
        parser.add_argument("-pn", "--parent_name", help="parent item name: e.g. project_name")
        parser.add_argument("-pt", "--parent_type", help="parent item type: e.g. projects")
        parser.add_argument("-ct", "--child_type", help=f"child item type: e.g. {item_type_deployment_processes}")
        parser.add_argument("-ck", "--child_id_key", help=f"child id key: e.g. {deployment_process_id_key}")
        parser.add_argument("-sk", "--sub_item_key",
                            help=f"sub-item key: this sub-item is for copy/clone e.g. {steps_key}")
        parser.add_argument("-bp", "--base_parent_name", help="base parent item name: e.g. a base project_name")
        parser.add_argument("-sn", "--step_name", help="step name as in octopus process")
        parser.add_argument("-bs", "--base_step_name", help="base step name as in octopus process for cloning a step")
        parser.add_argument("-ps", "--prev_step_name",
                            help="previous step name in octopus process for the step insertion location")
        parser.add_argument("-tl", "--time_limit_second", help="time limit in second")
        parser.add_argument("-rv", "--release_version", help="release version for creating a new release")
        parser.add_argument("-as", "--add_suffix",
                            help="if present, add suffix to variable sets name")
        parser.add_argument("-rs", "--remove_suffix", help="if present, remove suffix from variable sets name")
        parser.add_argument("-pj", "--project_name", help="project name")
        parser.add_argument("-cn", "--channel_name", help="channel name")
        parser.add_argument("-nt", "--notes", help="notes")
        parser.add_argument("-ri", "--release_id", help="release id for deployment")
        parser.add_argument("-en", "--environment_name", help="environment name, like Integration")
        parser.add_argument("-tn", "--tenant_name", help="tenant name, like cd-near")
        parser.add_argument("-cm", "--comments", help="comments")
        parser.add_argument("-eps", "--excluded_projects", help="comma delimited project names")
        parser.add_argument("-pgs", "--project_groups", help="comma delimited project group names")
        parser.add_argument("-pkg", "--package_history", help="if present, package_history = True", action="store_true")

        args, unknown = parser.parse_known_args()
        return args

    def _process_args_to_configs(self):
        args = self._parse_args()
        if args.action:
            self._target_config.action = args.action
        elif args.version:
            print(SystemConfig.TITLE)
            sys.exit()
        else:
            MainGUI().set_gui()
            sys.exit()

        if args.endpoint:
            self._target_config.endpoint = args.endpoint
        assert self._target_config.endpoint.endswith("/api/"), \
            f"octopus endpoint must end with /api/; {self._target_config.endpoint} is invalid"

        if args.data_path == Config.USE_CURRENT_DATA_PATH:
            self._target_config.data_path = os.getcwd()
        elif args.data_path:
            self._target_config.data_path = args.data_path

        if args.api_key and args.api_key.startswith("API-"):
            self._target_config.api_key = args.api_key
        elif args.api_key or args.api_key == "":
            self._target_common.log_info_print(
                msg=f"The octopus API-KEY does not start with 'API-'; so use user/password instead")
            self._target_config.api_key = ""

        if args.user_name:
            self._target_config.user_name = args.user_name
        if args.password:
            self._target_config.password = args.password
        assert self._target_config.api_key or (self._target_config.user_name and self._target_config.password), \
            f"either api_key or user_name and password are required"

        if args.pem and args.pem.lower() == "false":
            self._target_config.pem = False
        elif args.pem:
            self._target_config.pem = args.pem

        self._target_config.overwrite = args.overwrite
        logger.info(f"self._target_config.overwrite: {self._target_config.overwrite}")
        self._target_config.no_stdout = args.no_stdout
        logger.info(f"self._target_config.no_stdout: {self._target_config.no_stdout}")
        self._target_config.package_history = args.package_history
        logger.info(f"self._target_config.package_history: {self._target_config.package_history}")

        if args.space_id_name:
            self._target_config.space_id = self._target_common.verify_space(space_id_name=args.space_id_name)
            if self._target_config.space_id:
                logger.info(f"the target space_id is: {self._target_config.space_id}")
            else:
                raise ValueError(f"the space id/name {args.space_id_name} you specified does not exist or "
                                 f"you do not have permission to access it on server {self._target_config.endpoint}")

        if args.action != Actions.ACTION_GET_SPACES and not self._target_config.space_id \
                and input(f"Are you sure you want to run a command against {None} space [Y/n]? ") != "Y":
            return

        if args.action in MIGRATION_LIST:
            self._target_common.log_info_print(msg=f"===== Action: {args.action}; processing the source config =====")

            self._source_config.local_data = args.local_data
            logger.info(f"self._source_config.local_data: {self._source_config.local_data}")
            if args.source_endpoint:
                self._source_config.endpoint = args.source_endpoint
            assert self._source_config.local_data or self._source_config.endpoint.endswith("/api/"), \
                f"octopus endpoint must end with /api/; {self._source_config.endpoint} is invalid"

            if args.source_data_path == Config.USE_CURRENT_DATA_PATH:
                self._source_config.data_path = os.getcwd()
            elif args.source_data_path:
                self._source_config.data_path = args.source_data_path
            if args.local_data and self._target_config.endpoint != self._source_config.endpoint \
                    and self._target_config.data_path == self._source_config.data_path:
                raise ValueError(f"the source Octopus server {self._source_config.endpoint} and the target "
                                 f"Octopus server {self._target_config.endpoint} cannot use the same local "
                                 f"path {self._target_config.data_path} if cloning source is the local data")

            if args.source_api_key and args.source_api_key.startswith("API-"):
                self._source_config.api_key = args.source_api_key
            elif args.source_api_key or args.source_api_key == "":
                self._target_common.log_info_print(
                    msg=f"The source octopus API-KEY does not start with 'API-'; so use user/password instead")
                self._source_config.api_key = ""

            if args.source_user_name:
                self._source_config.user_name = args.source_user_name
            if args.source_password:
                self._source_config.password = args.source_password
            assert self._source_config.api_key or (self._source_config.user_name and self._source_config.password), \
                f"either api_key or user_name and password are required"

            if args.source_pem and args.source_pem.lower() == "false":
                self._source_config.pem = False
            elif args.source_pem:
                self._source_config.pem = args.source_pem

            if args.source_space_id_name:
                self._source_config.space_id = self._source_common.verify_space(space_id_name=args.source_space_id_name)
                if self._source_config.space_id:
                    self._target_common.log_info_print(msg=f"The source octopus space_id is: "
                                                           f"{self._source_config.space_id}")
                elif self._source_config.local_data:
                    self._target_common.log_info_print(
                        msg=f"{args.action} from local source data {args.source_space_id_name}")
                    self._source_config.space_id = args.source_space_id_name
            if not self._source_config.space_id and not self._source_config.local_data:
                raise ValueError(f"On Octopus server {self._source_config.endpoint}, the space id/name "
                                 f"{args.source_space_id_name} does not exist or you do not have permission to "
                                 f"access it.")

            if self._source_config.endpoint == self._target_config.endpoint:
                if args.action == Actions.ACTION_CLONE_SERVER:
                    raise ValueError(f"Cannot {args.action} from an endpoint to the same one: "
                                     f"{self._source_config.endpoint}")
                elif self._source_config.space_id == self._target_config.space_id:
                    raise ValueError(f"Cannot {args.action} from a space to the same space "
                                     f"{self._source_config.space_id} on the same Octopus server "
                                     f"{self._source_config.endpoint}")

            if args.action == Actions.ACTION_CLONE_SERVER:
                self._target_common.log_info_print(msg=f"{args.action} from {self._source_config.endpoint} to "
                                                       f"{self._target_config.endpoint}; space ids are cleared")
                self._source_config.space_id = None
                self._target_config.space_id = None
            elif not self._source_config.space_id or not self._target_config.space_id:
                raise ValueError(f"Cannot {args.action} from space {self._source_config.space_id} of the source "
                                 f"Octopus server {self._source_config.endpoint} to space "
                                 f"{self._target_config.space_id} of the target Octopus server "
                                 f"{self._target_config.endpoint}")

            self._source_config.save_config()

        self._target_config.save_config()

        return args

    def run(self):
        args = self._process_args_to_configs()

        if self._target_config.action == Actions.ACTION_GET_SPACES:
            self._target_common.get_spaces_save(item_types_comma_delimited=args.item_types,
                                                space_id_or_name_comma_delimited=args.spaces)
        elif self._target_config.action == Actions.ACTION_GET_TYPES:
            self._target_common.get_types_save(item_types_comma_delimited=args.item_types)
        elif self._target_config.action == Actions.ACTION_GET_TYPE:
            self._target_common.get_one_type_save(item_type=args.item_type)
        elif self._target_config.action == Actions.ACTION_DELETE_TYPE:
            self._target_common.delete_one_type(item_type=args.item_type)
        elif self._target_config.action == Actions.ACTION_DELETE_TYPES:
            self._target_common.delete_types(item_types_comma_delimited=args.item_types)
        elif self._target_config.action == Actions.ACTION_GET:
            self._target_common.get_single_item_by_name_or_id_save(item_type=args.item_type, item_name=args.item_name,
                                                                   item_id=args.item_id)
        elif self._target_config.action == Actions.ACTION_UPDATE:
            self._target_common.update_single_item_save(item_type=args.item_type, item_name=args.item_name,
                                                        item_id=args.item_id)
        elif self._target_config.action == Actions.ACTION_UPDATE_MERGE:
            self._target_common.merge_single_item_save(item_type=args.item_type, item_name=args.item_name,
                                                       item_id=args.item_id, child_id_key=args.child_id_key)
        elif self._target_config.action == Actions.ACTION_CREATE:
            self._target_common.create_single_item_from_local_file(item_type=args.item_type, item_name=args.item_name,
                                                                   local_item_name=args.base_item_name)
        elif self._target_config.action == Actions.ACTION_CLONE:
            self._target_common.clone_single_item_from_remote_item(item_type=args.item_type, item_name=args.item_name,
                                                                   base_item_name=args.base_item_name)
        elif self._target_config.action == Actions.ACTION_DELETE:
            self._target_common.delete_single_item_by_name_or_id(item_type=args.item_type, item_name=args.item_name,
                                                                 item_id=args.item_id)
        elif self._target_config.action == Actions.ACTION_GET_CHILD:
            self._target_common.get_child_item_save(parent_name=args.parent_name, parent_type=args.parent_type,
                                                    child_id_key=args.child_id_key, child_type=args.child_type)
        elif self._target_config.action == Actions.ACTION_UPDATE_CHILD:
            self._target_common.update_child_item_from_local_save(
                parent_name=args.parent_name, parent_type=args.parent_type, child_id_key=args.child_id_key,
                child_type=args.child_type)
        elif self._target_config.action == Actions.ACTION_CLONE_CHILD:
            self._target_common.clone_child_item_from_another_parent_save(
                parent_name=args.parent_name, base_parent_name=args.base_parent_name, parent_type=args.parent_type,
                child_id_key=args.child_id_key, child_type=args.child_type, sub_item_key=args.sub_item_key)
        elif self._target_config.action == Actions.ACTION_CLONE_PROCESS_STEP:
            DeploymentProcesses(config=self._target_config).clone_process_step(
                project_literal_name=args.project_name, step_name=args.step_name, base_step_name=args.base_step_name,
                prev_step_name=args.prev_step_name)
        elif self._target_config.action == Actions.ACTION_DELETE_PROCESS_STEP:
            DeploymentProcesses(config=self._target_config).delete_process_step(project_literal_name=args.project_name,
                                                                                step_name=args.step_name)
        elif self._target_config.action == Actions.ACTION_CLONE_PROJECT:
            Projects(config=self._target_config).clone_project(project_literal_name=args.project_name,
                                                               base_project_name=args.base_item_name)
        elif self._target_config.action == Actions.ACTION_DELETE_PROJECT:
            Projects(config=self._target_config).delete_project(project_literal_name=args.project_name)
        elif self._target_config.action == Actions.ACTION_DELETE_PROJECTS:
            Projects(config=self._target_config).delete_projects(
                project_groups_comma_delimited=args.project_groups,
                excluded_projects_comma_delimited=args.excluded_projects)
        elif self._target_config.action == Actions.ACTION_GET_PROJECT:
            Projects(config=self._target_config).get_project(project_literal_name=args.project_name)
        elif self._target_config.action == Actions.ACTION_PROJECT_UPDATE_VARIABLE_SETS:
            Projects(config=self._target_config).project_update_variable_sets(
                project_literal_name=args.project_name, remove_suffix=args.remove_suffix, add_suffix=args.add_suffix)
        elif self._target_config.action == Actions.ACTION_CLONE_SERVER:
            Migration(src_config=self._source_config, dst_config=self._target_config).clone_server(
                space_id_or_name_comma_delimited=args.spaces, item_types_comma_delimited=args.item_types)
        elif self._target_config.action == Actions.ACTION_CLONE_SPACE:
            Migration(src_config=self._source_config, dst_config=self._target_config).clone_space(
                item_types_comma_delimited=args.item_types)
        elif self._target_config.action == Actions.ACTION_CLONE_SPACE_ITEM:
            Migration(src_config=self._source_config, dst_config=self._target_config).clone_space_item(
                item_type=args.item_type, item_name=args.item_name, item_id=args.item_id)
        elif self._target_config.action == Actions.ACTION_TASK_STATUS:
            self._target_common.get_task_status(task_id=args.item_id)
        elif self._target_config.action == Actions.ACTION_WAIT_TASK:
            self._target_common.wait_task(task_id=args.item_id, time_limit_second=args.time_limit_second)
        elif self._target_config.action == Actions.ACTION_CREATE_RELEASE:
            ReleaseDeployment.create_release_direct(
                config=self._target_config, release_version=args.release_version, project_name=args.project_name,
                channel_name=args.channel_name, notes=args.notes)
        elif self._target_config.action == Actions.ACTION_CREATE_DEPLOYMENT:
            ReleaseDeployment.create_deployment_direct(
                config=self._target_config, release_id=args.release_id, environment_name=args.environment_name,
                tenant_name=args.tenant_name, comments=args.comments, project_name=args.project_name)
        elif self._target_config.action == Actions.ACTION_CREATE_RELEASE_DEPLOYMENT:
            ReleaseDeployment.create_release_deployment(
                config=self._target_config, release_version=args.release_version, project_name=args.project_name,
                channel_name=args.channel_name, notes=args.notes, environment_name=args.environment_name,
                tenant_name=args.tenant_name, comments=args.comments)
        elif self._target_config.action == Actions.ACTION_GUI:
            MainGUI().set_gui()
        else:
            log_raise_value_error(local_logger=logger, err="We only support actions: " + str(Actions.__dict__.values()))


def main():
    logger.info(f"********** {SystemConfig.TITLE} - start **********")
    OctopusClient().run()
    logger.info(f"********** Octopus deploy python client tool - done **********")


if __name__ == "__main__":
    main()



import getpass
import logging
import threading
import tkinter as tk
from pathlib import Path
from pprint import pformat
from time import strftime, localtime
from tkinter import messagebox

from octopus_python_client.actions import Actions, ACTIONS_DICT
from octopus_python_client.common import Common, item_type_channels, project_id_key, name_key, id_key, \
    item_type_projects, release_versions_key, version_key, item_type_packages, item_type_environments, item_type_tenants
from octopus_python_client.constants import Constants
from octopus_python_client.gui.common_widgets import CommonWidgets
from octopus_python_client.migration import Migration
from octopus_python_client.release_deployment import ReleaseDeployment
from octopus_python_client.utilities.helper import find_item

logger = logging.getLogger(__name__)


class SubmitWidgets(tk.Frame):
    DIVIDER_BAR = "|"
    TAS = "tas"

    def __init__(self, parent: tk.Frame, server: Common, source: Common, next_button: tk.Button = None,
                 submit_button: tk.Button = None):
        super().__init__(parent)
        self.server = server
        self.source = source

        self.next_button = next_button
        self.submit_button = submit_button

        self.channel_id_var = None
        self.combobox_var = None
        self.deployment_notes_var = None
        self.env_frame = tk.Frame(self)
        self.env_id_var = None
        self.local_file_var = None
        self.new_item_name_var = None
        self.new_name_entry = None
        self.overwrite_var = None
        self.package_history_var = None
        self.package_name = ""
        self.release_id = ""
        self.release_notes_var = None
        self.release_version_num_var = None
        self.services_versions_text = None
        self.services_versions_vs_name = ""
        self.tenant_frame = tk.Frame(self)
        self.tenant_id_var = None

        self.update_step()

    def update_step(self):
        tk.Label(self, text=f"{self.server.config.action} ({ACTIONS_DICT.get(self.server.config.action)})",
                 bd=2, relief="groove").grid(sticky=tk.W)

        if self.server.config.action == Actions.ACTION_CREATE_RELEASE:
            self.set_create_release_frame()

        elif self.server.config.action == Actions.ACTION_CREATE_DEPLOYMENT:
            self.set_create_deployment_frame()

        elif self.server.config.action == Actions.ACTION_CLONE_PROJECT_RELATED:
            if not self.server.config.project_ids:
                messagebox.showerror(title=f"No project selected", message=f"No destination project was selected!")
                self.submit_button.config(state=tk.DISABLED)
                return
            items_list = self.source.get_list_items_by_conditional_id(
                item_type=self.server.config.type, condition_key=project_id_key,
                condition_id=self.source.config.project_id)
            if self.assert_items_list(items_list=items_list):
                self.set_clone_item_frame(items_list=items_list)

        elif self.server.config.action == Actions.ACTION_CLONE_SPACE \
                or self.server.config.action == Actions.ACTION_GET_SPACES:
            if item_type_packages in self.server.config.types:
                self.set_package_history_widget()
            self.set_overwrite_widget()

        elif self.server.config.action == Actions.ACTION_CLONE_SPACE_ITEM:
            items_list = self.source.get_list_from_one_type(self.server.config.type)
            if self.assert_items_list(items_list=items_list):
                self.set_clone_item_frame(items_list=items_list)

        elif self.server.config.action == Actions.ACTION_GET:
            self.set_get_item_frame()

        elif self.server.config.action == Actions.ACTION_UPDATE:
            self.set_update_item_frame()

        else:
            self.submit_button.config(state=tk.DISABLED)

    def set_create_deployment_frame(self):
        releases_list = self.server.get_project_releases_sorted_list(project_id=self.server.config.project_id)
        project_name = self.server.get_item_name_by_id(item_type=item_type_projects,
                                                       item_id=self.server.config.project_id)
        if not releases_list:
            messagebox.showerror(
                title=f"No release",
                message=f"Project {project_name} does not have any releases. Please create a release first")
            self.submit_button.config(state=tk.DISABLED)
            return
        texts_list = [release.get(version_key) + SubmitWidgets.DIVIDER_BAR + release.get(id_key) for release in
                      releases_list]
        self.combobox_var = CommonWidgets.set_combobox_items_frame(
            parent=self, texts_list=texts_list, bind_func=self.process_release,
            title=f"Select a release for project {project_name} (the latest release is pre-selected)")
        self.deployment_notes_var = tk.StringVar()
        self.deployment_notes_var.set(self.server.config.deployment_notes)
        CommonWidgets.set_text_entry(parent=self, title="Deployment comments:", text_var=self.deployment_notes_var)
        self.process_release()

    def process_release(self, event=None):
        logger.info(msg=str(event))
        if SubmitWidgets.DIVIDER_BAR in self.combobox_var.get():
            release_version_id = self.combobox_var.get().split(SubmitWidgets.DIVIDER_BAR)
            self.release_id = release_version_id[1]
        else:
            self.release_id = self.combobox_var.get()
        promotion_information_dict = self.server.get_deployment_information(release_id=self.release_id)
        envs_list = promotion_information_dict.get("PromoteTo")
        tenants_list = promotion_information_dict.get("TenantPromotions")

        self.env_frame.destroy()
        self.tenant_frame.destroy()

        self.env_frame = tk.Frame(self)
        self.env_id_var = CommonWidgets.set_radio_items_frame(
            parent=self.env_frame, list_items=envs_list, title=f"Select an environment:")
        self.env_frame.grid(sticky=tk.W)

        self.tenant_frame = tk.Frame(self)
        self.tenant_id_var = CommonWidgets.set_radio_items_frame(
            parent=self.tenant_frame, list_items=tenants_list, default_id=self.server.config.tenant_id,
            title=f"Select a tenant:")
        self.tenant_frame.grid(sticky=tk.W)

    def set_release_notes(self, event=None):
        logger.info(msg=str(event))
        release_notes = "{'packages':{'" + self.package_name + "':'" + self.combobox_var.get() + \
                        "'},'release_versions':'" + self.services_versions_vs_name + "'}"
        self.release_notes_var.set(release_notes)
        self.server.config.package_version = self.combobox_var.get()
        self.update_services_versions()

    def set_services_versions_frame(self, project_name: str):
        tk.Label(self, text="Service versions pinned in Octopus variable set:").grid(sticky=tk.W)
        services_versions_vs_name_var = tk.StringVar()
        services_versions_vs_name_entry = tk.Entry(self, width=CommonWidgets.WIDTH_40,
                                                   textvariable=services_versions_vs_name_var)
        services_versions_vs_name_entry.grid(sticky=tk.W)
        services_versions_vs_name_entry.config(state=CommonWidgets.READ_ONLY)
        services_versions_vs_name_var.set(self.services_versions_vs_name)

        self.package_name = f"{SubmitWidgets.TAS}.{project_name}"
        package_history_list = self.server.get_package_history_list_by_name(package_name=self.package_name)
        package_versions_list = [package.get(version_key) for package in package_history_list]
        self.combobox_var = CommonWidgets.set_combobox_items_frame(
            parent=self, texts_list=package_versions_list, bind_func=self.set_release_notes,
            default_text=self.server.config.package_version, title=f"Select {self.package_name} version:",
            width=CommonWidgets.WIDTH_20)

    def set_release_notes_entry(self):
        tk.Label(self, text="Release notes (grayed out if the release notes is a python dictionary and used for pinning"
                            " package versions):").grid(sticky=tk.W)
        self.release_notes_var = tk.StringVar()
        # self.release_notes_var.set(self.server.config.release_notes)
        release_notes_entry = tk.Entry(self, width=CommonWidgets.WIDTH_120, textvariable=self.release_notes_var)
        release_notes_entry.grid(sticky=tk.W)
        return release_notes_entry

    def update_services_versions(self):
        self.services_versions_text.config(state=tk.NORMAL)
        self.services_versions_text.delete(1.0, tk.END)
        project_name = self.server.get_item_name_by_id(item_type=item_type_projects,
                                                       item_id=self.server.config.project_id)
        channel_name = self.server.get_item_name_by_id(item_type=item_type_channels, item_id=self.channel_id_var.get())
        select_packages_dict = ReleaseDeployment.get_package_versions(
            config=self.server.config, project_name=project_name, channel_name=channel_name,
            notes=self.release_notes_var.get())
        self.services_versions_text.insert(tk.END, pformat(select_packages_dict))
        self.services_versions_text.config(state=tk.DISABLED)

    def set_services_versions_textbox(self):
        tk.Label(self, text="The package versions which will be used for creating the new release:").grid(sticky=tk.W)
        self.services_versions_text = tk.Text(self, width=CommonWidgets.WIDTH_80, height=CommonWidgets.HEIGHT_7)
        self.services_versions_text.grid(sticky=tk.W)

    def set_create_release_frame(self):
        self.set_radio_channels_frame()

        tk.Label(self, text="Release version number: ").grid(sticky=tk.W)
        self.release_version_num_var = tk.StringVar()
        local_time = strftime("%Y.%m%d.%H%M%S", localtime())
        current_user = getpass.getuser()
        self.release_version_num_var.set(f"{local_time}-{current_user}")
        tk.Entry(self, width=CommonWidgets.WIDTH_40, textvariable=self.release_version_num_var).grid(sticky=tk.W)

        project_name = self.server.get_item_name_by_id(
            item_type=item_type_projects, item_id=self.server.config.project_id)
        self.services_versions_vs_name = release_versions_key + "." + project_name
        services_versions_list = self.server.get_list_variables_by_set_name_or_id(
            set_name=self.services_versions_vs_name)
        if services_versions_list:
            self.set_services_versions_frame(project_name=project_name)
            self.set_release_notes_entry().config(state=CommonWidgets.READ_ONLY)
            self.set_services_versions_textbox()
            self.set_release_notes()
        else:
            self.set_release_notes_entry()

    def set_get_item_frame(self):
        items_list = self.server.get_list_from_one_type(self.server.config.type)
        if self.assert_items_list(items_list=items_list):
            self.set_combobox_items_frame(items_list=items_list, common=self.server, bind_func=self.set_local_file)
            self.set_update_file_path_frame()
            self.set_local_file()
            self.set_check_package_overwrite()

    def set_update_item_frame(self):
        items_list = self.server.get_list_from_one_type(self.server.config.type)
        if self.assert_items_list(items_list=items_list):
            self.set_combobox_items_frame(items_list=items_list, common=self.server,
                                          bind_func=self.set_check_local_file)
            self.set_update_file_path_frame()
            self.set_check_local_file()

    def set_update_file_path_frame(self):
        tk.Label(self, text="The local file as the configuration").grid(sticky=tk.W)
        self.local_file_var = tk.StringVar()
        tk.Entry(self, width=CommonWidgets.WIDTH_120, textvariable=self.local_file_var, state=CommonWidgets.READ_ONLY) \
            .grid(sticky=tk.W)

    def assert_items_list(self, items_list: list):
        if not items_list or len(items_list) == 0:
            messagebox.showerror(title=f"No item", message=f"{self.server.config.type} has no item")
            self.submit_button.config(state=tk.DISABLED)
            return False
        return True

    def set_clone_item_frame(self, items_list: list):
        self.set_combobox_items_frame(items_list=items_list, common=self.source, bind_func=self.set_new_item_name)
        self.set_new_name_item_frame()
        self.set_new_item_name()
        self.set_check_package_overwrite()

    def set_check_package_overwrite(self):
        if item_type_packages == self.server.config.type:
            self.set_package_history_widget()
        self.set_overwrite_widget()

    def set_new_name_item_frame(self):
        tk.Label(self, text="New item name to be cloned (grayed out if the item has no name): ").grid(sticky=tk.W)
        self.new_item_name_var = tk.StringVar()
        self.new_name_entry = tk.Entry(self, width=CommonWidgets.WIDTH_40, textvariable=self.new_item_name_var)
        self.new_name_entry.grid(sticky=tk.W)

    def set_combobox_items_frame(self, items_list: list, common: Common, bind_func):
        default_item = find_item(lst=items_list, key=id_key, value=common.config.item_id)
        default_text = SubmitWidgets.construct_item_name_id_text(item=default_item)
        texts_list = [SubmitWidgets.construct_item_name_id_text(item=item) for item in items_list]
        self.combobox_var = CommonWidgets.set_combobox_items_frame(
            parent=self, texts_list=texts_list, bind_func=bind_func, default_text=default_text,
            title=f"Select an item for type {self.server.config.type} (item name|id, or version|id or id only:")

    def set_new_item_name(self, event=None):
        logger.info(msg=str(event))
        self.server.log_info_print(f"split {self.combobox_var.get()}")
        item_name, item_id = SubmitWidgets.split_item_name_id(self.combobox_var.get())
        if item_name:
            self.new_item_name_var.set(item_name)
        else:
            self.new_name_entry.config(state=CommonWidgets.READ_ONLY)
        self.source.config.item_id = item_id
        self.source.config.item_name = ""
        self.server.log_info_print(f"item_id: {item_id}")

    def set_check_local_file(self, event=None):
        logger.info(msg=str(event))
        local_file = self.set_local_file()
        self.check_local_file(local_file=local_file)

    def set_local_file(self, event=None):
        logger.info(msg=str(event))
        item_name, item_id = SubmitWidgets.split_item_name_id(self.combobox_var.get())
        self.server.config.item_id = item_id
        self.server.config.item_name = item_name
        local_file = self.server.get_local_single_item_file_smartly(
            item_type=self.server.config.type, item_name=item_name, item_id=item_id)
        self.local_file_var.set(local_file)
        return local_file

    def check_local_file(self, local_file):
        if Path(local_file).is_file():
            self.submit_button.config(state=tk.NORMAL)
        else:
            messagebox.showerror(title=f"File not exist", message=f"{local_file} does not exist")
            self.submit_button.config(state=tk.DISABLED)

    @staticmethod
    def split_item_name_id(item_name_id: str):
        if SubmitWidgets.DIVIDER_BAR in item_name_id:
            item_name_id_list = item_name_id.split(SubmitWidgets.DIVIDER_BAR)
            item_name = item_name_id_list[0]
            item_id = item_name_id_list[1]
        else:
            item_name = ""
            item_id = item_name_id
        return item_name, item_id

    @staticmethod
    def construct_item_name_id_text(item):
        if not item:
            return None
        elif isinstance(item, str):
            return item
        elif item.get(name_key) and item.get(id_key):
            return item.get(name_key) + SubmitWidgets.DIVIDER_BAR + item.get(id_key)
        elif item.get(id_key):
            return item.get(id_key)
        else:
            return ""

    def set_radio_channels_frame(self):
        channels_list = self.server.get_list_items_by_conditional_id(
            item_type=item_type_channels, condition_key=project_id_key, condition_id=self.server.config.project_id)
        default_channel = find_item(lst=channels_list, key=Constants.IS_DEFAULT, value=True)
        if default_channel:
            default_channel_id = default_channel.get(id_key)
        else:
            default_channel_id = self.server.config.channel_id
        self.channel_id_var = CommonWidgets.set_radio_items_frame(
            parent=self, list_items=channels_list, default_id=default_channel_id,
            title=f"Select a channel (default one is pre-selected):")

    def set_overwrite_widget(self):
        self.overwrite_var = tk.StringVar()
        tk.Checkbutton(self, text="Overwrite the existing entities of the same names, including the referenced entities"
                                  " (will skip if unchecked)", variable=self.overwrite_var).grid(sticky=tk.W)
        self.overwrite_var.set(CommonWidgets.SELECTED if self.server.config.overwrite else CommonWidgets.UNSELECTED)

    def set_package_history_widget(self):
        self.package_history_var = tk.StringVar()
        tk.Checkbutton(
            self, text="Clone/download all historical versions of the packages (clone/download only the latest version "
                       "if unchecked)", variable=self.package_history_var).grid(sticky=tk.W)
        self.package_history_var.set(
            CommonWidgets.SELECTED if self.server.config.package_history else CommonWidgets.UNSELECTED)

    def process_config(self):
        if self.overwrite_var:
            self.server.config.overwrite = True if self.overwrite_var.get() == CommonWidgets.SELECTED else False
        if self.channel_id_var:
            self.server.config.channel_id = self.channel_id_var.get()
        if self.release_version_num_var:
            self.server.config.release_version = self.release_version_num_var.get()
        if self.release_notes_var:
            self.server.config.release_notes = self.release_notes_var.get()
        if self.package_history_var:
            self.server.config.package_history = \
                True if self.package_history_var.get() == CommonWidgets.SELECTED else False
        if self.tenant_id_var:
            self.server.config.tenant_id = self.tenant_id_var.get()
        if self.deployment_notes_var:
            self.server.config.deployment_notes = self.deployment_notes_var.get()
        self.server.config.save_config()
        self.source.config.save_config()
        return True

    def run_thread(self):
        run_action = False
        historical_package_msg = f"\nThe historical versions of packages will " \
                                 f"{'' if self.server.config.package_history else 'NOT '}be downloaded. "
        overwrite_msg = f"\nThe existing entities will {'' if self.server.config.overwrite else 'NOT '}be overwritten."
        if self.server.config.action == Actions.ACTION_CLONE_SPACE:
            msg = f"Are you sure you want to clone types {self.server.config.types} from " \
                  f"{self.source.config.space_id} on server {self.source.config.endpoint} to " \
                  f"{self.server.config.space_id} on server {self.server.config.endpoint}?"
            msg += (historical_package_msg if item_type_packages in self.server.config.types else "")
            msg += overwrite_msg
            if messagebox.askyesno(title=f"{self.server.config.action}", message=msg):
                run_action = True
                Migration(src_config=self.source.config, dst_config=self.server.config).clone_space_types()

        elif self.server.config.action == Actions.ACTION_CLONE_SPACE_ITEM:
            # msg = f"cloning {item_type} {item_badge} from {self._src_config.space_id} on server "
            # f"{self._src_config.endpoint} to {self._dst_config.space_id} on server {self._dst_config.endpoint}")
            msg = f"Are you sure you want to clone type {self.server.config.type} of {self.combobox_var.get()}" \
                  f" from {self.source.config.space_id} on server {self.source.config.endpoint} to " \
                  f"{self.new_item_name_var.get()} in {self.server.config.space_id} on server " \
                  f"{self.server.config.endpoint}?"
            msg += (historical_package_msg if item_type_packages == self.server.config.type else "")
            msg += overwrite_msg
            if messagebox.askyesno(title=f"{self.server.config.action}", message=msg):
                run_action = True
                pars_dict = None
                if self.new_item_name_var.get():
                    pars_dict = {Constants.NEW_ITEM_NAME_KEY: self.new_item_name_var.get()}
                Migration(src_config=self.source.config, dst_config=self.server.config) \
                    .clone_space_item_new_name(pars_dict=pars_dict)

        elif self.server.config.action == Actions.ACTION_CLONE_PROJECT_RELATED:
            msg = f"Are you sure you want to clone type {self.server.config.type} of {self.combobox_var.get()}" \
                  f" from {self.source.config.space_id} on server {self.source.config.endpoint} to " \
                  f"{self.new_item_name_var.get()} in projects {self.server.config.project_ids} in " \
                  f"{self.server.config.space_id} on server {self.server.config.endpoint}?"
            msg += overwrite_msg
            if messagebox.askyesno(title=f"{self.server.config.action}", message=msg):
                run_action = True
                pars_dict = {Constants.NEW_ITEM_NAME_KEY: self.new_item_name_var.get(),
                             Constants.PROJECT_IDS_KEY: self.server.config.project_ids}
                Migration(src_config=self.source.config, dst_config=self.server.config) \
                    .clone_space_item_new_name(pars_dict=pars_dict)

        elif self.server.config.action == Actions.ACTION_CREATE_RELEASE:
            project_name = self.server.get_or_delete_single_item_by_id(
                item_type=item_type_projects, item_id=self.server.config.project_id).get(name_key)
            channel_name = self.server.get_or_delete_single_item_by_id(
                item_type=item_type_channels, item_id=self.server.config.channel_id).get(name_key)
            msg = f"Are you sure you want to create a new release for project {project_name} with release version " \
                  f"{self.server.config.release_version}, and channel {channel_name}, release notes " \
                  f"{self.server.config.release_notes}?"
            if messagebox.askyesno(title=f"{self.server.config.action}", message=msg):
                run_action = True
                ReleaseDeployment.create_release_direct(
                    config=self.server.config, release_version=self.server.config.release_version,
                    project_name=project_name, channel_name=channel_name, notes=self.server.config.release_notes)

        elif self.server.config.action == Actions.ACTION_CREATE_DEPLOYMENT:
            project_name = self.server.get_item_name_by_id(
                item_type=item_type_projects, item_id=self.server.config.project_id)
            env_name = self.server.get_item_name_by_id(
                item_type=item_type_environments, item_id=self.env_id_var.get())
            tenant_name = self.server.get_item_name_by_id(
                item_type=item_type_tenants, item_id=self.tenant_id_var.get())
            msg = f"Are you sure you want to create a new deployment for {self.release_id} in project {project_name} " \
                  f"with environment {env_name}, tenant {tenant_name} and comments {self.deployment_notes_var.get()}?"
            if messagebox.askyesno(title=f"{self.server.config.action}", message=msg):
                run_action = True
                ReleaseDeployment.create_deployment_direct(
                    config=self.server.config, environment_name=env_name, tenant_name=tenant_name,
                    release_id=self.release_id, project_name=project_name, comments=self.deployment_notes_var.get())

        elif self.server.config.action == Actions.ACTION_GET_SPACES:
            msg = f"Are you sure you want to download types {self.server.config.types} from spaces " \
                  f"{self.server.config.space_ids} on server {self.server.config.endpoint}?"
            msg += (historical_package_msg if item_type_packages in self.server.config.types else "")
            msg += overwrite_msg
            if messagebox.askyesno(title=f"{self.server.config.action}", message=msg):
                run_action = True
                spaces_ids = ",".join(self.server.config.space_ids)
                types = ",".join(self.server.config.types)
                self.server.get_spaces_save(space_id_or_name_comma_delimited=spaces_ids,
                                            item_types_comma_delimited=types)

        elif self.server.config.action == Actions.ACTION_GET:
            msg = f"Are you sure you want to download item {self.server.config.item_id} from space " \
                  f"{self.server.config.space_id} on server {self.server.config.endpoint} to local file " \
                  f"{self.local_file_var.get()}?"
            msg += (historical_package_msg if item_type_packages == self.server.config.type else "")
            msg += overwrite_msg
            if messagebox.askyesno(title=f"{self.server.config.action}", message=msg):
                run_action = True
                self.server.get_single_item_by_name_or_id_save(
                    item_type=self.server.config.type, item_id=self.server.config.item_id)

        elif self.server.config.action == Actions.ACTION_UPDATE:
            msg = f"Are you sure you want to update item {self.server.config.item_name} {self.server.config.item_id} " \
                  f"from local file {self.local_file_var.get()} to space {self.server.config.space_id} on server " \
                  f"{self.server.config.endpoint}?"
            if messagebox.askyesno(title=f"{self.server.config.action}", message=msg):
                run_action = True
                self.server.update_single_item_save(
                    item_type=self.server.config.type, item_name=self.server.config.item_name,
                    item_id=self.server.config.item_id)

        else:
            self.server.log_info_print("not a valid action")

        if run_action:
            messagebox.showinfo(
                title="Done!",
                message=f"{self.server.config.action} ({ACTIONS_DICT.get(self.server.config.action)}) is done!")

    def start_run(self):
        threading.Thread(target=self.run_thread).start()



class Constants:
    CHANNELS_ITEM_TYPE = "channels"
    IS_DEFAULT = "IsDefault"
    LIBRARY_NAME_KEY = "library_name"
    NEW_ITEM_NAME_KEY = "new_item_name"
    PACKAGE_NAME_KEY = "package_name"
    PACKAGE_VERSION_KEY = "package_version"
    PROJECT_IDS_KEY = "project_ids"
    PROJECT_TRIGGERS_ITEM_TYPE = "projecttriggers"
    RUNBOOKS_ITEM_TYPE = "runbooks"
    TRIGGERS_ITEM_TYPE = "triggers"

    PROJECT_RELATED_TYPES = [CHANNELS_ITEM_TYPE, RUNBOOKS_ITEM_TYPE, PROJECT_TRIGGERS_ITEM_TYPE]



import copy
import json
import logging
from os import getcwd
from os.path import dirname, abspath, join
from pathlib import Path

from octopus_python_client.actions import Actions
from octopus_python_client.constants import Constants
from octopus_python_client.utilities.get_pypi_version import get_version
from octopus_python_client.utilities.helper import load_file, save_file

logging.basicConfig(filename=join(getcwd(), "octopus_python_client.log"),
                    filemode="a",
                    format="%(asctime)s,%(msecs)d %(name)s %(levelname)s %(message)s",
                    datefmt="%H:%M:%S",
                    level=logging.INFO)


class SystemConfig:
    with open(f"{dirname(abspath(__file__))}/configurations/system_config.json") as fp:
        system_config = json.load(fp)
    PACKAGE_VERSION = system_config.get(Constants.PACKAGE_VERSION_KEY)
    PACKAGE_NAME = system_config.get(Constants.PACKAGE_NAME_KEY)
    LIBRARY_NAME = system_config.get(Constants.LIBRARY_NAME_KEY)
    LATEST_PYPI_VERSION = get_version(PACKAGE_NAME)
    TITLE = f"Octopus Python Client - Developed by Tony Li & Tableau Cloud Engineering Team - Version " \
            f"{PACKAGE_VERSION} - the latest PYPI version is {LATEST_PYPI_VERSION}"


class BaseConfig:
    API_KEY = "api_key"
    ENDPOINT = "endpoint"
    PASSWORD = "password"
    USER_NAME = "user_name"
    DATA_PATH = "data_path"

    def __init__(self, is_source_server: bool = False):
        self.api_key = ""
        self.data_path = getcwd()
        self.endpoint = ""
        self.is_source_server = is_source_server
        self.item_id = ""
        self.item_name = ""
        self.local_data = False
        self.password = ""
        self.pem = False
        self.project_id = ""
        self.space_id = ""
        self.user_name = ""
        self._base_config_dict = copy.deepcopy(self.__dict__)


class Config(BaseConfig):
    _BASE_CONFIG_DICT_KEY = "_base_config_dict"
    CONFIG_FILE_KEY = "config_file"
    CONFIGURATIONS_FOLDER = "configurations"
    DEFAULT_CONFIGURATION_FILE_NAME = "configuration.json"
    USE_CURRENT_DATA_PATH = "current"
    LOGGER = logging.getLogger("Config")
    SOURCE_SERVER_JSON = "source_server.json"

    def __init__(self, configuration_file_name: str = None, is_source_server: bool = False):
        super().__init__(is_source_server=is_source_server)

        self.action = Actions.ACTION_GET_SPACES
        self.channel_id = ""
        self.deployment_notes = ""
        self.no_stdout = False
        self.overwrite = False
        self.package_history = False
        self.package_version = ""
        self.prev_releases = []
        self.project_ids = []
        self.release_notes = ""
        self.release_version = ""
        self.space_ids = []
        self.tenant_id = ""
        self.type = ""
        self.types = []

        configuration_file_name = configuration_file_name if configuration_file_name else \
            (Config.SOURCE_SERVER_JSON if is_source_server else Config.DEFAULT_CONFIGURATION_FILE_NAME)
        code_path = dirname(abspath(__file__))
        self.config_file = join(code_path, Config.CONFIGURATIONS_FOLDER, configuration_file_name)
        self.load_config()

    def load_config(self):
        if Path(self.config_file).is_file():
            Config.LOGGER.info(f"loading configuration from {self.config_file}...")
            config_dict = load_file(self.config_file)
            data_path = self.data_path
            self.__dict__.update(config_dict)
            if not self.data_path:
                self.data_path = data_path
        else:
            Config.LOGGER.info(f"configuration file {self.config_file} does not exist.")

    def save_config(self):
        if self.is_source_server:
            self._base_config_dict.update(
                (k, self.__dict__[k]) for k in self._base_config_dict.keys() & self.__dict__.keys())
            save_file(file_path_name=self.config_file, content=self._base_config_dict)
        else:
            config_dict = copy.deepcopy(self.__dict__)
            config_dict.pop(Config.CONFIG_FILE_KEY, None)
            config_dict.pop(Config._BASE_CONFIG_DICT_KEY, None)
            save_file(file_path_name=self.config_file, content=config_dict)


if __name__ == "__main__":
    print(Config.LOGGER.name)
    source_config = Config(configuration_file_name="source_server.json", is_source_server=True)
    print(source_config.__dict__)
    source_config.save_config()
    main_config = Config()
    main_config.save_config()
